/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 */

#ifndef __CARD_H
#define __CARD_H

#include "compil.h"

typedef struct {
	/* Horizontal CRTC */
	int HTotal;
	int HDisp;
	int HBStart;
	int HSStart;
	int HSEnd;
	int HBEnd;

	/* Vertical CRTC, doubled in doublescan, halved in interlaced */
	int VTotal;
	int VDisp;
	int VBStart;
	int VSStart;
	int VSEnd;
	int VBEnd;

	int width; /* Size of the mode */
	int height;
	int original_width; /* Original size of the mode */
	int original_height;

	int hpolarity;
	int vpolarity;
	int doublescan; /* Doublescan flag */
	int interlace; /* Interlace flag */
	int interlaceratio; /* Interlace ratio, 0-100 */
	int bits; /* Bits per pixel */

	/* Clocks */
	long dotclockHz;
} card_crtc;

void card_crtc_all_set(card_crtc STACK_PTR *cp);
void card_generic_all_set(card_crtc STACK_PTR *cp);
DWORD card_bitmov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos);

void card_out(int edx, BYTE al);
BYTE card_in(int edx);
void card_outw(int edx, WORD ax);
WORD card_inw(int edx);
void card_outl(int edx, DWORD eax);
DWORD card_inl(int edx);

int card_crtc_address_get(void);
BYTE card_seq_get(BYTE);
void card_seq_set(BYTE, BYTE);
BYTE card_crtc_get(BYTE);
void card_crtc_set(BYTE, BYTE);
BYTE card_attr_get(BYTE);
void card_attr_set(BYTE, BYTE);
BYTE card_graph_get(BYTE);
void card_graph_set(BYTE, BYTE);
int card_divider_get(void);
void card_divider_set(BYTE);
void card_polarity_set(int HPolar, int VPolar);
void card_doublescan_set(int flag);
void card_signal_disable(void);
void card_signal_enable(void);
void card_unlock(void);
void card_lock(void);
void card_char_size_x_set(unsigned);
void card_char_size_y_set(unsigned);
void card_pll_set(unsigned);
unsigned card_pll_get(void);
long card_clock_compute(long dotclock, int mul_min, int mul_max, int div_min, int div_max, int p_min, int p_max, long ref, long vco_min, long vco_max, int STACK_PTR* s_mul, int STACK_PTR* s_div, int STACK_PTR* s_p, int (*validate)(int,int,int));

/***************************************************************************/
/* Log */

#ifdef __PCIREAL__

#define CARD_LOG(a) do { } while (0)

#else

void card_log(const char *text,...);

#define CARD_LOG(a) card_log a

#endif

/***************************************************************************/
/* PCI */

int pci_detect(void);
int pci_find_device(unsigned vendor, unsigned device, unsigned STACK_PTR* bus_device_func);
int pci_read_dword(unsigned bus_device_func, unsigned reg, DWORD STACK_PTR* value);
int pci_write_dword(unsigned bus_device_func, unsigned reg, DWORD value);
int pci_scan_device(void (*callback)(unsigned bus_device_func,unsigned vendor,unsigned device));

int pci_BIOS_address_map(unsigned bus_device_func);
void pci_BIOS_address_unmap(void);
void pci_BIOS_read(void STACK_PTR* dest, unsigned offset, unsigned len);

int pci_MMIO_address_map(unsigned bus_device_func, unsigned reg, DWORD mask);
void pci_MMIO_address_unmap(void);
int pci_MMIO_address_relocate(void);
void pci_MMIO_address_restore(void);
unsigned pci_MMIO_selector_get(void);

/***************************************************************************/
/* Mapping */

extern int card_create_linear_mapping(unsigned long *linear, unsigned long physaddr, int size);
extern int card_create_selector(unsigned *segment, unsigned long linear, int size);
extern void card_remove_linear_mapping(unsigned long *linear);
extern void card_remove_selector(unsigned *segment);

#endif
