
/*
 * Configuration routines.
 *
 * 19971219 support for mame.cfg by Valerio Verrando
 * 19980402 moved out of msdos.c (N.S.), generalized routines (BW)
 * 19980917 added a "-cheatfile" option (misc) in MAME.CFG      JCK
 */

#include "mamalleg.h"
#include "driver.h"
#include "screen/screen.h"
#include "screen/scrblit.h"
#include <ctype.h>

/* from main() */
extern int ignorecfg;

#include "hscript.h"
#include "video.h"
#include "record.h"

/* from sound.c */
extern int soundcard, usestereo, attenuation, sampleratedetect;

/* from input.c */
extern int use_mouse, joystick, use_hotrod;
extern unsigned idle_limit;

/* from cheat.c */
extern char *cheatfile;

/* from datafile.c */
extern const char *history_filename,*mameinfo_filename;

/* from fileio.c */
void decompose_rom_sample_path (const char *rompath, const char *samplepath);

#ifdef MESS
void decompose_software_path (char *softwarepath);
#endif

extern const char *nvdir, *hidir, *cfgdir, *inpdir, *stadir, *memcarddir;
extern const char *artworkdir, *screenshotdir;
extern char *alternate_name;

extern const char *cheatdir;

#ifdef MESS
/* path to the CRC database files */
char *crcdir;
char *softwarepath;  /* for use in fileio.c */
#define CHEAT_NAME	"CHEAT.CDB"
#define HISTRY_NAME "SYSINFO.DAT"
#else
#define CHEAT_NAME	"CHEAT.DAT"
#define HISTRY_NAME "HISTORY.DAT"
#endif

static int mame_argc;
static char **mame_argv;
static int game;
const char *rompath, *samplepath;

struct { char *name; int id; } joy_table[] =
{
	{ "none",           JOY_TYPE_NONE },
	{ "auto",           JOY_TYPE_AUTODETECT },
	{ "standard",       JOY_TYPE_STANDARD },
	{ "dual",           JOY_TYPE_2PADS },
	{ "4button",        JOY_TYPE_4BUTTON },
	{ "6button",        JOY_TYPE_6BUTTON },
	{ "8button",        JOY_TYPE_8BUTTON },
	{ "fspro",          JOY_TYPE_FSPRO },
	{ "wingex",         JOY_TYPE_WINGEX },
	{ "sidewinder",     JOY_TYPE_SIDEWINDER_AG },
	{ "gamepadpro",     JOY_TYPE_GAMEPAD_PRO },
	{ "grip",           JOY_TYPE_GRIP },
	{ "grip4",          JOY_TYPE_GRIP4 },
	{ "sneslpt1",       JOY_TYPE_SNESPAD_LPT1 },
	{ "sneslpt2",       JOY_TYPE_SNESPAD_LPT2 },
	{ "sneslpt3",       JOY_TYPE_SNESPAD_LPT3 },
	{ "psxlpt1",        JOY_TYPE_PSXPAD_LPT1 },
	{ "psxlpt2",        JOY_TYPE_PSXPAD_LPT2 },
	{ "psxlpt3",        JOY_TYPE_PSXPAD_LPT3 },
	{ "n64lpt1",        JOY_TYPE_N64PAD_LPT1 },
	{ "n64lpt2",        JOY_TYPE_N64PAD_LPT2 },
	{ "n64lpt3",        JOY_TYPE_N64PAD_LPT3 },
	{ "wingwarrior",    JOY_TYPE_WINGWARRIOR },
	{ "segaisa",        JOY_TYPE_IFSEGA_ISA },
	{ "segapci",        JOY_TYPE_IFSEGA_PCI },
	{ 0, 0 }
} ;

/* Convert the pair [section],option in a sub option name for the game specific sections */
char* get_sub_option(const char* section, const char* option) {
	static char buffer[256];
	if (strcmp(section,"config")==0) {
		sprintf(buffer,"%s",option);
	} else {
		sprintf(buffer,"%s_%s",section,option);
	}
	return buffer;
}

/*
 * gets some boolean config value.
 * 0 = false, >0 = true, <0 = auto
 * the shortcut can only be used on the commandline
 */
static int get_bool (char *section, char *option, char *shortcut, int def)
{
	const char *yesnoauto;
	int res, i;

	res = def;

	if (ignorecfg) goto cmdline;

	/* look into mame.cfg, [section] */
	if (def == 0)
		yesnoauto = get_config_string(section, option, "no");
	else if (def > 0)
		yesnoauto = get_config_string(section, option, "yes");
	else /* def < 0 */
		yesnoauto = get_config_string(section, option, "auto");

	/* if the option doesn't exist in the cfgfile, create it */
	if (get_config_string(section, option, "#") == "#")
		set_config_string(section, option, yesnoauto);

	/* look into mame.cfg, [resolution] */
	yesnoauto = get_config_string(get_game_resolution(drivers[game]), get_sub_option(section,option), yesnoauto);

	/* look into mame.cfg, [gamename] */
	yesnoauto = get_config_string((char *)drivers[game]->name, get_sub_option(section,option), yesnoauto);

	/* also take numerical values instead of "yes", "no" and "auto" */
	if		(stricmp(yesnoauto, "no"  ) == 0) res = 0;
	else if (stricmp(yesnoauto, "yes" ) == 0) res = 1;
	else if (stricmp(yesnoauto, "auto") == 0) res = -1;
	else	res = atoi (yesnoauto);

cmdline:
	/* check the commandline */
	for (i = 1; i < mame_argc; i++)
	{
		if (mame_argv[i][0] != '-') continue;
		/* look for "-option" */
		if (stricmp(&mame_argv[i][1], option) == 0)
			res = 1;
		/* look for "-shortcut" */
		if (shortcut && (stricmp(&mame_argv[i][1], shortcut) == 0))
			res = 1;
		/* look for "-nooption" */
		if (strnicmp(&mame_argv[i][1], "no", 2) == 0)
		{
			if (stricmp(&mame_argv[i][3], option) == 0)
				res = 0;
			if (shortcut && (stricmp(&mame_argv[i][3], shortcut) == 0))
				res = 0;
		}
		/* look for "-autooption" */
		if (strnicmp(&mame_argv[i][1], "auto", 4) == 0)
		{
			if (stricmp(&mame_argv[i][5], option) == 0)
				res = -1;
			if (shortcut && (stricmp(&mame_argv[i][5], shortcut) == 0))
				res = -1;
		}
	}
	return res;
}

static int get_int (char *section, char *option, char *shortcut, int def)
{
	int res,i;

	res = def;

	if (!ignorecfg)
	{
		/* if the option does not exist, create it */
		if (get_config_int (section, option, -777) == -777)
			set_config_int (section, option, def);

		/* look into mame.cfg, [section] */
		res = get_config_int (section, option, def);

		/* look into mame.cfg, [resolution] */
		res = get_config_int (get_game_resolution(drivers[game]), get_sub_option(section,option), res);

		/* look into mame.cfg, [gamename] */
		res = get_config_int ((char *)drivers[game]->name, get_sub_option(section,option), res);
	}

	/* get it from the commandline */
	for (i = 1; i < mame_argc; i++)
	{
		if (mame_argv[i][0] != '-')
			continue;
		if ((stricmp(&mame_argv[i][1], option) == 0) ||
			(shortcut && (stricmp(&mame_argv[i][1], shortcut ) == 0)))
		{
			i++;
			if (i < mame_argc) res = atoi (mame_argv[i]);
		}
	}
	return res;
}

static float get_float (char *section, char *option, char *shortcut, float def)
{
	int i;
	float res;

	res = def;

	if (!ignorecfg)
	{
		/* if the option does not exist, create it */
		if (get_config_float (section, option, 9999.0) == 9999.0)
			set_config_float (section, option, def);

		/* look into mame.cfg, [section] */
		res = get_config_float (section, option, def);

		/* look into mame.cfg, [resolution] */
		res = get_config_float (get_game_resolution(drivers[game]), get_sub_option(section,option), res);

		/* look into mame.cfg, [gamename] */
		res = get_config_float ((char *)drivers[game]->name, get_sub_option(section,option), res);
	}

	/* get it from the commandline */
	for (i = 1; i < mame_argc; i++)
	{
		if (mame_argv[i][0] != '-')
			continue;
		if ((stricmp(&mame_argv[i][1], option) == 0) ||
			(shortcut && (stricmp(&mame_argv[i][1], shortcut ) == 0)))
		{
			i++;
			if (i < mame_argc) res = atof (mame_argv[i]);
		}
	}
	return res;
}

static const char *get_string (char *section, char *option, char *shortcut, char *def)
{
	const char *res;
	int i;

	res = def;

	if (!ignorecfg)
	{
		/* if the option does not exist, create it */
		if (get_config_string (section, option, "#") == "#" )
			set_config_string (section, option, def);

		/* look into mame.cfg, [section] */
		res = get_config_string(section, option, def);

		/* look into mame.cfg, [resolution] */
		res = get_config_string(get_game_resolution(drivers[game]), get_sub_option(section,option), res);

		/* look into mame.cfg, [gamename] */
		res = get_config_string((char*)drivers[game]->name, get_sub_option(section,option), res);
	}

	/* get it from the commandline */
	for (i = 1; i < mame_argc; i++)
	{
		if (mame_argv[i][0] != '-')
			continue;

		if ((stricmp(&mame_argv[i][1], option) == 0) ||
			(shortcut && (stricmp(&mame_argv[i][1], shortcut)  == 0)))
		{
			i++;
			if (i < mame_argc) res = mame_argv[i];
		}
	}
	return res;
}

void get_rom_sample_path (int argc, char **argv, int game_index, char *override_default_rompath)
{
	int i;

	alternate_name = 0;
	mame_argc = argc;
	mame_argv = argv;
	game = game_index;

	rompath = override_default_rompath;
	if (rompath == NULL || rompath[0] == 0)
	{
#ifndef MESS
		rompath    = get_string ("directory", "rompath",    NULL, ".;ROMS");
#else
		rompath    = get_string ("directory", "biospath",   NULL, ".;BIOS");
#endif
	}

#ifndef MESS
	samplepath = get_string ("directory", "samplepath", NULL, ".;SAMPLES");
#else
	softwarepath= get_string ("directory", "softwarepath", NULL, ".;SOFTWARE");
#endif

	/* handle '-romdir' hack. We should get rid of this BW */
	alternate_name = 0;
	for (i = 1; i < argc; i++)
	{
		if (stricmp (argv[i], "-romdir") == 0)
		{
			i++;
			if (i < argc) alternate_name = argv[i];
		}
	}

	/* decompose paths into components (handled by fileio.c) */
	decompose_rom_sample_path (rompath, samplepath);
}


/* for playback of .inp files */
void init_inpdir(void)
{
	inpdir = get_string ("directory", "inp",     NULL, "INP");
}

/* Compute the blit orientation to mimic the rol/ror/flipx/flipy variables */
static int orientation_compute(int rol, int ror, int flipx, int flipy) {
	int orientation = ROT0;
	if (ror)
	{
		/* if only one of the components is inverted, switch them */
		if ((orientation & ROT180) == ORIENTATION_FLIP_X ||
				(orientation & ROT180) == ORIENTATION_FLIP_Y)
			orientation ^= ROT180;

		orientation ^= ROT90;
	}
	if (rol)
	{
		/* if only one of the components is inverted, switch them */
		if ((orientation & ROT180) == ORIENTATION_FLIP_X ||
				(orientation & ROT180) == ORIENTATION_FLIP_Y)
			orientation ^= ROT180;

		orientation ^= ROT270;
	}
	if (flipx)
	{
		orientation ^= ORIENTATION_FLIP_X;
	}
	if (flipy)
	{
		orientation ^= ORIENTATION_FLIP_Y;
	}
	return orientation;
}

const char* GAME_BLIT_ROTATE[] = {
#include "blitrot.h"
0
};

int is_game_in_list(const char* list[], const struct GameDriver* pgame) {
	while (pgame && pgame->name) {
		unsigned i;
		for(i=0;list[i];++i) {
			if (stricmp(pgame->name,list[i])==0)
				return 1;
		}
		pgame = pgame->clone_of;
	}
	return 0;
}


int parse_cmdline (int argc, char **argv, int game_index, char *override_default_rompath)
{
	static float f_beam, f_flicker;
	const char *joyname;
	int i;
	const char *tmpstr;

	mame_argc = argc;
	mame_argv = argv;
	game = game_index;


	/* force third mouse button emulation to "no" otherwise Allegro will default to "yes" */
	set_config_string(0,"emulate_three","no");

	/* read graphic configuration */
	scanlines	= get_bool	 ("config", "scanlines",    NULL,  1);
	options.use_artwork = get_bool	 ("config", "artwork",  NULL,  1);
	options.use_samples = get_bool	 ("config", "samples",  NULL,  1);
	video_vsync_request = get_bool   ("config", "vsync",        NULL,  0);
	video_vsync_factor = get_float ("config", "vsync_correction_factor",   NULL, 0.99);
	video_vsync_adjust_request = get_bool   ("config", "vsync_adjust",        NULL,  0);
	wait_vsync_request = get_bool   ("config", "waitvsync",    NULL,  0);
	use_triplebuf_request = get_bool ("config", "triplebuffer",	NULL,  0);
	dirty_request = get_bool	 ("config", "dirty",	NULL,	-1);
	options.antialias	= get_bool	 ("config", "antialias",    NULL,  1);
	options.translucency = get_bool    ("config", "translucency", NULL, 1);
	tmpstr = get_string ("config", "resize", NULL,  "mixed");
	if (stricmp(tmpstr,"integer")==0) {
		stretch_request = STRETCH_INTEGER_XY;
	} else if (stricmp(tmpstr,"fractional")==0) {
		stretch_request = STRETCH_FRACTIONAL_XY;
	} else if (stricmp(tmpstr,"mixed")==0) {
		stretch_request = STRETCH_INTEGER_X_FRACTIONAL_Y;
	} else if (stricmp(tmpstr,"no")==0 || stricmp(tmpstr,"none")==0) {
		stretch_request = STRETCH_NONE;
	} else {
		fprintf(stderr,"Invalid option resize = %s\n",tmpstr);
		fprintf(stderr,"Valid values are: no, integer, mixed, fractional\n");
		return 1;
	}
	tmpstr = get_string ("config", "aspect", NULL, "fit_used");
	if (stricmp(tmpstr,"square")==0) {
		aspect_request = ASPECT_SQUARE;
	} else if (stricmp(tmpstr,"fit_used")==0) {
		aspect_request = ASPECT_FIT_USED;;
	} else if (stricmp(tmpstr,"fit_whole")==0) {
		aspect_request = ASPECT_FIT_WHOLE;
	} else {
		fprintf(stderr,"Invalid option aspect = %s\n",tmpstr);
		fprintf(stderr,"Valid values are: square, fit_used, fit_whole\n");
		return 1;
	}

	tmpstr			   = get_string ("config", "depth", NULL, "auto");
	options.color_depth = atoi(tmpstr);
	if (options.color_depth != 8 && options.color_depth != 15 &&
		options.color_depth != 16 && options.color_depth != 32)
		options.color_depth = 0; /* auto */

	tmpstr = get_string ("config", "skiplines", NULL,  "auto");
	if (stricmp(tmpstr,"auto")==0) {
		skiplines_request = -1;
	} else {
		skiplines_request = atoi(tmpstr);
	}
	tmpstr = get_string ("config", "skipcolumns", NULL,  "auto");
	if (stricmp(tmpstr,"auto")==0) {
		skipcolumns_request = -1;
	} else {
		skipcolumns_request = atoi(tmpstr);
	}
	f_beam		= get_float  ("config", "beam",         NULL, 1.0);
	if (f_beam < 1.0) f_beam = 1.0;
	if (f_beam > 16.0) f_beam = 16.0;
	f_flicker	= get_float  ("config", "flicker",      NULL, 0.0);
	if (f_flicker < 0.0) f_flicker = 0.0;
	if (f_flicker > 100.0) f_flicker = 100.0;
	osd_gamma_correction = get_float ("config", "gamma",   NULL, 1.0);
	if (osd_gamma_correction < 0.5) osd_gamma_correction = 0.5;
	if (osd_gamma_correction > 2.0) osd_gamma_correction = 2.0;
	brightness = 100 * get_float ("config", "brightness",   NULL, 1.0);

	tmpstr			   = get_string ("config", "frameskip", "fs", "auto");
	if (!stricmp(tmpstr,"auto"))
	{
		frameskip_request = 0;
		autoframeskip_request = 1;
	}
	else
	{
		frameskip_request = atoi(tmpstr);
		autoframeskip_request = 0;
	}
	options.norotate  = get_bool ("config", "norotate",  NULL, 0);
	options.ror 	  = get_bool ("config", "ror",       NULL, 0);
	options.rol 	  = get_bool ("config", "rol",       NULL, 0);
	options.flipx	  = get_bool ("config", "flipx",     NULL, 0);
	options.flipy	  = get_bool ("config", "flipy",     NULL, 0);

	tmpstr = get_string ("config", "blitrotate", NULL, "auto");
	if (stricmp(tmpstr,"auto")==0) {
		if (is_game_in_list(GAME_BLIT_ROTATE,drivers[game_index]))
			tmpstr = "yes";
		else
			tmpstr = "no";
	}
	if (stricmp(tmpstr,"yes")==0 && (drivers[game_index]->drv->video_attributes & VIDEO_TYPE_VECTOR)!=0) {
		logerror("warning: blitrotate forced to no because the game is a vector game\n");
		tmpstr = "no";
	}
	if (stricmp(tmpstr,"yes")==0) {
		blit_orientation = orientation_compute(options.rol,options.ror,options.flipx,options.flipy);
		options.ror = 0;
		options.rol = 0;
		options.flipx = 0;
		options.flipy = 0;
	} else if (stricmp(tmpstr,"no")==0) {
		blit_orientation = ROT0;
	} else {
		fprintf(stderr,"Invalid option blitrotate = %s, valid values are auto, yes, no\n",tmpstr);
		return 1;
	}

	tmpstr = get_string ("config", "bliteffect", NULL, "auto");
	if (stricmp(tmpstr,"auto")==0) {
		blit_combine_request = COMBINE_AUTO;
	} else if (stricmp(tmpstr,"max")==0) {
		blit_combine_request = COMBINE_MAX;
	} else if (stricmp(tmpstr,"mean")==0) {
		blit_combine_request = COMBINE_MEAN;
	} else if (stricmp(tmpstr,"filter")==0) {
		blit_combine_request = COMBINE_FILTER;
	} else if (stricmp(tmpstr,"filterx")==0) {
		blit_combine_request = COMBINE_FILTERX;
	} else if (stricmp(tmpstr,"filtery")==0) {
		blit_combine_request = COMBINE_FILTERY;
	} else if (stricmp(tmpstr,"no")==0 || stricmp(tmpstr,"none")==0) {
		blit_combine_request = COMBINE_NONE;
	} else {
		fprintf(stderr,"Invalid option bliteffect = %s\n",tmpstr);
		fprintf(stderr,"Valid values are: auto, no, max, mean, filter\n");
		return 1;
	}

	/* read sound configuration */
	sound_record_time = get_int  ("config", "recordtime",  NULL, 8);
	soundcard			= get_int  ("config", "soundcard",  NULL, -1);
	options.use_emulated_ym3812 = !get_bool ("config", "ym3812opl",  NULL,  0);
	options.samplerate = get_int  ("config", "samplerate", "sr", 44100);
	if (options.samplerate < 5000) options.samplerate = 5000;
	if (options.samplerate > 50000) options.samplerate = 50000;
	usestereo			= get_bool ("config", "stereo",  NULL,  1);
	attenuation 		= get_int  ("config", "volume",  NULL,  0);
	if (attenuation < -32) attenuation = -32;
	if (attenuation > 0) attenuation = 0;
	sampleratedetect    = get_bool ("config", "sampleratedetect", NULL, 1);
	options.use_filter = get_bool ("config", "resamplefilter", NULL, 1);

	/* read input configuration */
	use_mouse = get_bool   ("config", "mouse",   NULL,  1);
	joyname   = get_string ("config", "joystick", "joy", "none");
	use_hotrod = 0;
	if (get_bool ("config", "hotrod",   NULL,  0)) use_hotrod = 1;
	if (get_bool ("config", "hotrodse",   NULL,  0)) use_hotrod = 2;
	idle_limit = get_int("config","idleexit",NULL,0);

	/* turbo configuration */
	turbo_speed = get_float("config", "turbospeed", NULL, 3.0);
	if (turbo_speed < 1.0) turbo_speed = 1.0;
	if (turbo_speed > 30.0) turbo_speed = 30.0;
	turbo_startup_time = get_int("config","turbostartuptime",NULL,6);
	if (turbo_startup_time < 0) turbo_startup_time = 0;
	if (turbo_startup_time > 180) turbo_startup_time = 180;

	/* misc configuration */
	options.cheat	   = get_bool ("config", "cheat", NULL, 0);
#ifdef MAME_DEBUG
	options.mame_debug = 0;
#endif

	tmpstr = get_string ("config", "cheatfile", "cf", CHEAT_NAME );
	cheatfile = malloc(strlen(tmpstr) + 1);
	strcpy(cheatfile,tmpstr);

	history_filename  = get_string ("config", "historyfile", NULL, HISTRY_NAME);
	mameinfo_filename  = get_string ("config", "mameinfofile", NULL, "MAMEINFO.DAT");    /* JCK 980917 */

	/* get resolution */
	tmpstr = get_string ("config", "video", NULL, "auto");
	if (tmpstr && *tmpstr)
		strcpy(resolution_request,tmpstr);
	else
		strcpy(resolution_request,"auto");
	video_mode_reset_at_exit = get_bool ("config", "video_mode_reset", NULL, 1);

	/* get video config compare */
	tmpstr = get_string ("config", "video_factor_game_visible", NULL, "8.0");
	gfx_cmp.game_visible = atof(tmpstr);
	tmpstr = get_string ("config", "video_factor_game_stretch", NULL, "4.0");
	gfx_cmp.game_stretch = atof(tmpstr);
	tmpstr = get_string ("config", "video_factor_screen_visible", NULL, "1.0");
	gfx_cmp.screen_visible = atof(tmpstr);
	tmpstr = get_string ("config", "video_factor_mode_size", NULL, "1.0");
	gfx_cmp.mode_size = atof(tmpstr);
	tmpstr = get_string ("config", "video_factor_frequency", NULL, "0.0");
	gfx_cmp.frequency = atof(tmpstr);
	tmpstr = get_string ("config", "video_factor_aspect", NULL, "1.0");
	gfx_cmp.aspect = atof(tmpstr);

	/* set default subdirectories */
#ifndef MESS
	hidir		= get_string ("directory", "hi",      NULL, "HI");
	stadir		= get_string ("directory", "sta",     NULL, "STA");
	cheatdir	= get_string ("directory",  "cheat",   NULL, ".");
#else
	crcdir		= get_string ("directory",  "crc",     NULL, "CRC");
	cheatdir	= get_string ("directory",  "cheat",   NULL, "CHEAT");
#endif
	memcarddir	= get_string ("directory", "memcard", NULL, "MEMCARD");
	nvdir		= get_string ("directory", "nvram",   NULL, "NVRAM");
	artworkdir	= get_string ("directory", "artwork", NULL, "ARTWORK");
	cfgdir		= get_string ("directory", "cfg",     NULL, "CFG");
	screenshotdir = get_string ("directory", "snap", NULL, "SNAP");


	logerror("cheatfile = %s - cheatdir = %s\n",cheatfile,cheatdir);

	tmpstr = get_string ("config", "language", NULL, "english");
	options.language_file = osd_fopen(0,tmpstr,OSD_FILETYPE_LANGUAGE,0);

	/* this is handled externally cause the audit stuff needs it, too */
	get_rom_sample_path (argc, argv, game_index, override_default_rompath);

	/* process some parameters */
	options.beam = (int)(f_beam * 0x00010000);
	if (options.beam < 0x00010000)
		options.beam = 0x00010000;
	if (options.beam > 0x00100000)
		options.beam = 0x00100000;

	options.flicker = (int)(f_flicker * 2.55);
	if (options.flicker < 0)
		options.flicker = 0;
	if (options.flicker > 255)
		options.flicker = 255;

	/* convert joystick name into an Allegro-compliant joystick signature */
	joystick = -2; /* default to invalid value */

	for (i = 0; joy_table[i].name != NULL; i++)
	{
		if (stricmp (joy_table[i].name, joyname) == 0)
		{
			joystick = joy_table[i].id;
			logerror("using joystick %s = %08x\n", joyname,joy_table[i].id);
			break;
		}
	}

	if (joystick == -2)
	{
		logerror("%s is not a valid entry for a joystick\n", joyname);
		joystick = JOY_TYPE_NONE;
	}

#ifdef MESS
	/* Is there an override for softwarepath= for this driver? */
	tmpstr = get_string ((char*)drivers[game]->name, "softwarepath", NULL, NULL);
	if (tmpstr)
		softwarepath = tmpstr;
	logerror("Using Software Path %s for %s\n", softwarepath, drivers[game]->name);
	decompose_software_path(softwarepath);
#endif

	hardware_script_config(get_string);

	return 0;
}

