#include "hscript.h"
#include "script.h"

#include <mamalleg.h>
#include <stdio.h>
#include <string.h>

/* Current script event state */
static int the_script_condition;

/* Script source */
static char* the_hardware_script_text_map[HARDWARE_SCRIPT_MAX];

/* Set the source */
void hardware_script_set(int id, const char* script) {
	free(the_hardware_script_text_map[id]);
	if (script && *script)
		the_hardware_script_text_map[id] = strdup(script);
	else
		the_hardware_script_text_map[id] = 0;
}

struct hardware_script_pair {
	int condition;
	struct script_cmd* script;
	struct script_state* state;
};

static struct hardware_script_pair the_hardware_script_map[HARDWARE_SCRIPT_MAX];

static unsigned char the_kdb_state;

unsigned the_kdb_flags[3] = {
	KB_NUMLOCK_FLAG,
	KB_CAPSLOCK_FLAG,
	KB_SCROLOCK_FLAG
};

/* Port callback */
unsigned char script_port_read(int address) {
	if (address == 0) {
		return the_kdb_state;
	} else
		return inportb(address);
}

void script_port_write(int address,unsigned char value) {
	if (address == 0) {
		value &= 0x7;
		if (the_kdb_state != value) {
			unsigned led_mask = 0;
			if (value & 0x1) led_mask |= the_kdb_flags[0];
			if (value & 0x2) led_mask |= the_kdb_flags[1];
			if (value & 0x4) led_mask |= the_kdb_flags[2];
			set_leds(led_mask);
			the_kdb_state = value;
		}
	} else {
		outportb(address,value);
	}
}

int script_symbol_check(const char* sym) {
	if (strcmp(sym,"lpt1")==0) {
		return 1;
	} else if (strcmp(sym,"kdb")==0) {
		return 1;
	}
	return 0;
}

int script_symbol_get(const char* sym) {
	if (strcmp(sym,"lpt1")==0) {
		return 0x378;
	} else if (strcmp(sym,"kdb")==0) {
		return 0;
	}
	return 0;
}

int script_function_check(const char* sym, int argc) {
	if (argc == 0 && strcmp(sym,"event")==0) {
		return 1;
	} else if (argc == 1 && strcmp(sym,"get")==0) {
		return 1;
	} else if (argc == 2 && strcmp(sym,"set")==0) {
		return 1;
	} else if (argc == 2 && strcmp(sym,"on")==0) {
		return 1;
	} else if (argc == 2 && strcmp(sym,"off")==0) {
		return 1;
	} else if (argc == 2 && strcmp(sym,"toggle")==0) {
		return 1;
	}
	return 0;
}

int script_function0_get(const char* sym) {
	if (strcmp(sym,"event")==0) {
		return the_script_condition;
	}
	return 0;
}

int script_function1_get(const char* sym, int arg0) {
	if (strcmp(sym,"get")==0) {
		return script_port_read(arg0);
	}
	return 0;
}

int script_function2_get(const char* sym, int arg0, int arg1) {
	if (strcmp(sym,"set")==0) {
		script_port_write(arg0,arg1);
		return 0;
	} else if (strcmp(sym,"on")==0) {
		script_port_write(arg0, script_port_read(arg0) | arg1);
		return 0;
	} else if (strcmp(sym,"off")==0) {
		script_port_write(arg0, script_port_read(arg0) & ~arg1);
		return 0;
	} else if (strcmp(sym,"toggle")==0) {
		script_port_write(arg0, script_port_read(arg0) ^ arg1);
		return 0;
	}
	return 0;
}

static const char* the_script_text;

/* Parse error callback */
void script_error(const char* s) {
	fprintf(stderr,"Error compiling the script: %s\n",the_script_text);
	fprintf(stderr,"%s\n",s);
}

int hardware_script_init(void) {
	int i;
	for(i=0;i<HARDWARE_SCRIPT_MAX;++i) {
		if (the_hardware_script_text_map[i]) {
			the_script_text = the_hardware_script_text_map[i];
			the_hardware_script_map[i].script = script_parse(the_hardware_script_text_map[i]);
			if (!the_hardware_script_map[i].script)
				return -1;
			the_hardware_script_map[i].state = script_run_alloc();
		} else {
			the_hardware_script_map[i].script = 0;
			the_hardware_script_map[i].state = 0;
		}
	}
	return 0;
}

void hardware_script_done(void) {
	int i;
	for(i=0;i<HARDWARE_SCRIPT_MAX;++i) {
		script_free(the_hardware_script_map[i].script);
		script_run_free(the_hardware_script_map[i].state);
		free(the_hardware_script_text_map[i]);
	}
}

/* Start the script */
void hardware_script_start(int id) {
	if (the_hardware_script_map[id].script) {
		the_hardware_script_map[id].condition = 1;
		/* start only if not already running */
		if (script_run_end(the_hardware_script_map[id].state))
			script_run_restart(the_hardware_script_map[id].state,the_hardware_script_map[id].script);
	}
}

void hardware_script_stop(int id) {
	if (the_hardware_script_map[id].script) {
		the_hardware_script_map[id].condition = 0;
	}
}

static void hardware_script_idle_single(unsigned id, unsigned time_to_play) {
	if (the_hardware_script_map[id].script) {
		the_script_condition = the_hardware_script_map[id].condition;
		script_run(the_hardware_script_map[id].state,time_to_play);
	}
}

/* Use idle time */
void hardware_script_idle(unsigned time_to_play) {
	int i;
	for(i=0;i<HARDWARE_SCRIPT_MAX;++i) {
		hardware_script_idle_single(i,time_to_play);
	}
}

void hardware_script_terminate(int id) {
	int time = SCRIPT_TIME_UNIT; /* one second time to flush any delay */
	hardware_script_stop(id);
	hardware_script_idle_single(id,time);
}

void hardware_script_config(const char* (*get_string)(char *section, char *option, char *shortcut, char *def)) {
	const char* tmpstr;

	tmpstr = get_string ("script", "video", NULL, "wait(!event()); set(kdb,0);");
	hardware_script_set(HARDWARE_SCRIPT_VIDEO, tmpstr);

	tmpstr = get_string ("script", "emulation", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_EMULATION, tmpstr);

	tmpstr = get_string ("script", "play", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_PLAY, tmpstr);

	tmpstr = get_string ("script", "led1", NULL, "on(kdb, 0b1); wait(!event()); off(kdb, 0b1);");
	hardware_script_set(HARDWARE_SCRIPT_LED1, tmpstr);

	tmpstr = get_string ("script", "led2", NULL, "on(kdb, 0b10); wait(!event()); off(kdb, 0b10);");
	hardware_script_set(HARDWARE_SCRIPT_LED2, tmpstr);

	tmpstr = get_string ("script", "led3", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_LED3, tmpstr);

	tmpstr = get_string ("script", "coin1", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_COIN1, tmpstr);

	tmpstr = get_string ("script", "coin2", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_COIN2, tmpstr);

	tmpstr = get_string ("script", "coin3", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_COIN3, tmpstr);

	tmpstr = get_string ("script", "coin4", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_COIN4, tmpstr);

	tmpstr = get_string ("script", "start1", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_START1, tmpstr);

	tmpstr = get_string ("script", "start2", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_START2, tmpstr);

	tmpstr = get_string ("script", "start3", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_START3, tmpstr);

	tmpstr = get_string ("script", "start4", NULL, "");
	hardware_script_set(HARDWARE_SCRIPT_START4, tmpstr);

	tmpstr = get_string ("script", "turbo", NULL, "while (event()) { toggle(kdb, 0b100); delay(100); } off(kdb, 0b100);");
	hardware_script_set(HARDWARE_SCRIPT_TURBO, tmpstr);
}
