/***************************************************************************/
/* Video mode container extension */

static int mode_cmp(const video_mode* a,const video_mode* b) {
	int av;
	int bv;

	if (a->size_y < b->size_y)
		return -1;
	if (a->size_y > b->size_y)
		return 1;
	if (a->size_x < b->size_x)
		return -1;
	if (a->size_x > b->size_x)
		return 1;

	av = a->flags & VIDEO_FLAGS_DRIVER_MASK;
	bv = b->flags & VIDEO_FLAGS_DRIVER_MASK;
	if (av < bv)
		return -1;
	if (av > bv)
		return 1;

	if (a->bits < b->bits)
		return -1;
	if (a->bits > b->bits)
		return 1;

	av = a->flags & VIDEO_FLAGS_LINEAR;
	bv = b->flags & VIDEO_FLAGS_LINEAR;
	if (av > bv)
		return -1;
	if (av < bv)
		return 1;


	return 0;
}

static video_mode* video_mode_container_pos(video_mode_container* vmc, unsigned pos) {
	video_mode_container_iterator i;
	video_mode_container_iterator_begin(&i,vmc);
	while (pos && !video_mode_container_iterator_end(&i)) {
		--pos;
		video_mode_container_iterator_next(&i);
	}
	if (video_mode_container_iterator_end(&i))
		return 0;
	else
		return video_mode_container_iterator_get(&i);
}

static unsigned video_mode_container_max(video_mode_container* vmc) {
	unsigned pos = 0;
	video_mode_container_iterator i;
	video_mode_container_iterator_begin(&i,vmc);
	while (!video_mode_container_iterator_end(&i)) {
		++pos;
		video_mode_container_iterator_next(&i);
	}
	return pos;
}

/***************************************************************************/
/* Sound */

int the_sound_flag = 1;

static void sound_error(void) {
	if (the_sound_flag) {
		sound(100);
		delay(120);
		nosound();
	}
}

static void sound_warn(void) {
	if (the_sound_flag) {
		sound(900);
		delay(10);
		nosound();
	}
}

static void sound_signal(void) {
	if (the_sound_flag) {
		unsigned i;
		for(i=0;i<10;++i) {
			sound(800);
			delay(10);
			nosound();
			delay(10);
		}
	}
}

/***************************************************************************/
/* Monitor */

static video_monitor the_monitor;

static int monitor_check(const video_mode* mode) {
	return video_monitor_mode_check(&the_monitor, mode);
}

static int monitor_check_safe(const video_mode* mode) {
	return video_monitor_mode_check_safe(&the_monitor, mode);
}

/***************************************************************************/
/* Text output */

#define DEFAULT_TEXT_MODE "default_text"

static struct bitmapfont* the_font;
static video_mode_container the_modes; /* main video mode container */
static video_mode the_default_mode;  /* Default video mode */
static int the_default_mode_flag = 0; /* Default video mode set */

static unsigned screen_size_x(void) {
	return video_size_x() / video_font_size_x();
}

static unsigned screen_size_y(void) {
	return video_size_y() / video_font_size_y();
}

static void screen_clear(void) {
	unsigned x,y;
	for(x=0;x<screen_size_x();++x)
		for(y=0;y<screen_size_y();++y)
			video_put_char(x,y,0,0);
}

static void screen_set_font(void) {
	if (video_is_text()) {
		if (the_font) {
			bitmapfont_set(the_font);
		} else {
			unsigned font_row = video_size_y() / 25;
			if (font_row < 14)
				vga_font_set_default8();
			else if (font_row < 16)
				vga_font_set_default14();
			else
				vga_font_set_default16();
		}
	}
}

/* Smaller is better */
static int screen_video_mode_compare(const video_mode* a, const video_mode* b) {
	int a_check;
	int b_check;
	if (strcmp(a->name,DEFAULT_TEXT_MODE)==0 && strcmp(b->name,DEFAULT_TEXT_MODE)!=0)
		return -1;
	if (strcmp(a->name,DEFAULT_TEXT_MODE)!=0 && strcmp(b->name,DEFAULT_TEXT_MODE)==0)
		return 1;
	if ((a->flags & VIDEO_FLAGS_USER_BIT0)!=0 && (b->flags & VIDEO_FLAGS_USER_BIT0)==0)
		return -1;
	if ((a->flags & VIDEO_FLAGS_USER_BIT0)==0 && (b->flags & VIDEO_FLAGS_USER_BIT0)!=0)
		return 1;
	a_check = (a->flags & VIDEO_FLAGS_USER_BIT0)!=0 ? monitor_check(a) : monitor_check_safe(a);
	b_check = (b->flags & VIDEO_FLAGS_USER_BIT0)!=0 ? monitor_check(b) : monitor_check_safe(b);
	if (a_check && !b_check)
		return -1;
	if (!a_check && b_check)
		return 1;
	a_check = a->size_x * a->size_y;
	b_check = b->size_x * b->size_y;
	return b_check - a_check;
}

static int screen_default_set(void) {
	video_mode_container_iterator i;

	/* search the best mode */
	for(video_mode_container_iterator_begin(&i,&the_modes);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if (video_mode_is_text(vm)) {
			if (!the_default_mode_flag || screen_video_mode_compare(vm,&the_default_mode)<0) {
				the_default_mode = *vm;
				the_default_mode_flag = 1;
			}
		}
	}

	/* use the current */
	if (!the_default_mode_flag) {
		if (vgaline_mode_grab(&the_default_mode.driver.vgaline)==0 &&
			vgaline_mode_import(&the_default_mode)==0 &&
			video_mode_is_text(&the_default_mode)) {
			strcpy(the_default_mode.name,DEFAULT_TEXT_MODE);
			video_mode_container_insert_sort(&the_modes,&the_default_mode,mode_cmp);
			the_default_mode_flag = 1;
		}
	}

	if (!the_default_mode_flag)
		return 0;

	if (video_mode_set(&the_default_mode)!=0)
		return -1;

	screen_set_font();
	return 0;
}

static void screen_init(void) {
	the_font = bitmapfont_load("mv.fnt");
#if 0
	if (!the_font) {
		video_mode_reset();
		printf("Error loading the font file <mv.fnt>\n");
		exit(EXIT_FAILURE);
	}
#endif

	if (screen_default_set()!=0) {
		video_mode_reset();
		printf("Error inizialing the default video mode\n\"%s\"",video_error_description_get());
		exit(EXIT_FAILURE);
	}

	if (the_default_mode_flag == 0) {
		video_mode_reset();
		printf("No video mode avaliable for your hardware\n");
		exit(EXIT_FAILURE);
	}

}

static void screen_reset(void) {
	if (video_mode_active())
		video_mode_done();
	if (video_mode_set(&the_default_mode)!=0) {
		video_mode_reset();
		printf("Error in inizialing the default video mode\n");
		exit(EXIT_FAILURE);
	}
	screen_set_font();
}

static void screen_done(void) {
	if (the_font)
		bitmapfont_free(the_font);

	video_mode_reset();
}

static void screen_put(int x, int y, char c, int color) {
	if (y>=0 && y<screen_size_y() && x>=0 && x<screen_size_x()) {
		video_put_char(x,y,c,color);
	}
}

/**************************************************************************/
/* Input */

#define INPUT_KEY_UP 18432
#define INPUT_KEY_DOWN 20480
#define INPUT_KEY_LEFT 19200
#define INPUT_KEY_RIGHT 19712
#define INPUT_KEY_CTRLUP 36096
#define INPUT_KEY_CTRLDOWN 37120
#define INPUT_KEY_CTRLRIGHT 29696
#define INPUT_KEY_CTRLLEFT 29440
#define INPUT_KEY_ENTER 13
#define INPUT_KEY_ESC 27
#define INPUT_KEY_Y 121
#define INPUT_KEY_HOME 18176
#define INPUT_KEY_END 20224
#define INPUT_KEY_PGUP 18688
#define INPUT_KEY_PGDN 20736
#define INPUT_KEY_F1 15104
#define INPUT_KEY_F2 15360
#define INPUT_KEY_F3 15616
#define INPUT_KEY_F4 15872
#define INPUT_KEY_F5 16128
#define INPUT_KEY_F6 16384
#define INPUT_KEY_F7 16640
#define INPUT_KEY_F8 16896
#define INPUT_KEY_F9 17152
#define INPUT_KEY_F10 17408
#define INPUT_KEY_BACKSPACE 8
#define INPUT_KEY_DEL 21248
#define INPUT_KEY_INS 20992
#define INPUT_KEY_SPACE ' '
#define INPUT_KEY_TAB '\t'

static int input_kbhit(void) {
	return kbhit();
}

static unsigned input_getkey(void) {
	char c = getch();
	if (!c) {
		unsigned e = getch();
		return e << 8;
	} else
		return c;
}

/***************************************************************************/
/* Draw interface */

#define FONTX 6
#define FONTY 8

static unsigned char draw_font[] = {
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x30,0x48,0x84,0xb4,0xb4,0x84,0x48,0x30,0x30,0x48,0x84,0x84,0x84,0x84,0x48,0x30,
		0x00,0xfc,0x84,0x8c,0xd4,0xa4,0xfc,0x00,0x00,0xfc,0x84,0x84,0x84,0x84,0xfc,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x30,0x68,0x78,0x78,0x30,0x00,0x00,
		0x80,0xc0,0xe0,0xf0,0xe0,0xc0,0x80,0x00,0x04,0x0c,0x1c,0x3c,0x1c,0x0c,0x04,0x00,
		0x20,0x70,0xf8,0x20,0x20,0xf8,0x70,0x20,0x48,0x48,0x48,0x48,0x48,0x00,0x48,0x00,
		0x00,0x00,0x30,0x68,0x78,0x30,0x00,0x00,0x00,0x30,0x68,0x78,0x78,0x30,0x00,0x00,
		0x70,0xd8,0xe8,0xe8,0xf8,0xf8,0x70,0x00,0x1c,0x7c,0x74,0x44,0x44,0x4c,0xcc,0xc0,
		0x20,0x70,0xf8,0x70,0x70,0x70,0x70,0x00,0x70,0x70,0x70,0x70,0xf8,0x70,0x20,0x00,
		0x00,0x10,0xf8,0xfc,0xf8,0x10,0x00,0x00,0x00,0x20,0x7c,0xfc,0x7c,0x20,0x00,0x00,
		0xb0,0x54,0xb8,0xb8,0x54,0xb0,0x00,0x00,0x00,0x28,0x6c,0xfc,0x6c,0x28,0x00,0x00,
		0x00,0x30,0x30,0x78,0x78,0xfc,0x00,0x00,0xfc,0x78,0x78,0x30,0x30,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x20,0x20,0x20,0x20,0x20,0x00,0x20,0x00,
		0x50,0x50,0x50,0x00,0x00,0x00,0x00,0x00,0x00,0x50,0xf8,0x50,0xf8,0x50,0x00,0x00,
		0x20,0x70,0xc0,0x70,0x18,0xf0,0x20,0x00,0x40,0xa4,0x48,0x10,0x20,0x48,0x94,0x08,
		0x60,0x90,0xa0,0x40,0xa8,0x90,0x68,0x00,0x10,0x20,0x40,0x00,0x00,0x00,0x00,0x00,
		0x20,0x40,0x40,0x40,0x40,0x40,0x20,0x00,0x10,0x08,0x08,0x08,0x08,0x08,0x10,0x00,
		0x20,0xa8,0x70,0xf8,0x70,0xa8,0x20,0x00,0x00,0x20,0x20,0xf8,0x20,0x20,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x60,0x00,0x00,0x00,0xf8,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x00,0x00,0x08,0x10,0x20,0x40,0x80,0x00,0x00,
		0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x00,0x10,0x30,0x10,0x10,0x10,0x10,0x10,0x00,
		0x70,0x88,0x08,0x10,0x20,0x40,0xf8,0x00,0x70,0x88,0x08,0x30,0x08,0x88,0x70,0x00,
		0x10,0x30,0x50,0x90,0xf8,0x10,0x10,0x00,0xf8,0x80,0xf0,0x08,0x08,0x88,0x70,0x00,
		0x70,0x80,0xf0,0x88,0x88,0x88,0x70,0x00,0xf8,0x08,0x08,0x10,0x20,0x20,0x20,0x00,
		0x70,0x88,0x88,0x70,0x88,0x88,0x70,0x00,0x70,0x88,0x88,0x88,0x78,0x08,0x70,0x00,
		0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x00,0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x60,
		0x10,0x20,0x40,0x80,0x40,0x20,0x10,0x00,0x00,0x00,0xf8,0x00,0xf8,0x00,0x00,0x00,
		0x40,0x20,0x10,0x08,0x10,0x20,0x40,0x00,0x70,0x88,0x08,0x10,0x20,0x00,0x20,0x00,
		0x30,0x48,0x94,0xa4,0xa4,0x94,0x48,0x30,0x70,0x88,0x88,0xf8,0x88,0x88,0x88,0x00,
		0xf0,0x88,0x88,0xf0,0x88,0x88,0xf0,0x00,0x70,0x88,0x80,0x80,0x80,0x88,0x70,0x00,
		0xf0,0x88,0x88,0x88,0x88,0x88,0xf0,0x00,0xf8,0x80,0x80,0xf0,0x80,0x80,0xf8,0x00,
		0xf8,0x80,0x80,0xf0,0x80,0x80,0x80,0x00,0x70,0x88,0x80,0x98,0x88,0x88,0x70,0x00,
		0x88,0x88,0x88,0xf8,0x88,0x88,0x88,0x00,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x00,
		0x08,0x08,0x08,0x08,0x88,0x88,0x70,0x00,0x88,0x90,0xa0,0xc0,0xa0,0x90,0x88,0x00,
		0x80,0x80,0x80,0x80,0x80,0x80,0xf8,0x00,0x88,0xd8,0xa8,0x88,0x88,0x88,0x88,0x00,
		0x88,0xc8,0xa8,0x98,0x88,0x88,0x88,0x00,0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x00,
		0xf0,0x88,0x88,0xf0,0x80,0x80,0x80,0x00,0x70,0x88,0x88,0x88,0x88,0x88,0x70,0x08,
		0xf0,0x88,0x88,0xf0,0x88,0x88,0x88,0x00,0x70,0x88,0x80,0x70,0x08,0x88,0x70,0x00,
		0xf8,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x88,0x88,0x88,0x88,0x88,0x88,0x70,0x00,
		0x88,0x88,0x88,0x88,0x88,0x50,0x20,0x00,0x88,0x88,0x88,0x88,0xa8,0xd8,0x88,0x00,
		0x88,0x50,0x20,0x20,0x20,0x50,0x88,0x00,0x88,0x88,0x88,0x50,0x20,0x20,0x20,0x00,
		0xf8,0x08,0x10,0x20,0x40,0x80,0xf8,0x00,0x30,0x20,0x20,0x20,0x20,0x20,0x30,0x00,
		0x40,0x40,0x20,0x20,0x10,0x10,0x08,0x08,0x30,0x10,0x10,0x10,0x10,0x10,0x30,0x00,
		0x20,0x50,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xfc,
		0x40,0x20,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x70,0x08,0x78,0x88,0x78,0x00,
		0x80,0x80,0xf0,0x88,0x88,0x88,0xf0,0x00,0x00,0x00,0x70,0x88,0x80,0x80,0x78,0x00,
		0x08,0x08,0x78,0x88,0x88,0x88,0x78,0x00,0x00,0x00,0x70,0x88,0xf8,0x80,0x78,0x00,
		0x18,0x20,0x70,0x20,0x20,0x20,0x20,0x00,0x00,0x00,0x78,0x88,0x88,0x78,0x08,0x70,
		0x80,0x80,0xf0,0x88,0x88,0x88,0x88,0x00,0x20,0x00,0x20,0x20,0x20,0x20,0x20,0x00,
		0x20,0x00,0x20,0x20,0x20,0x20,0x20,0xc0,0x80,0x80,0x90,0xa0,0xe0,0x90,0x88,0x00,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x00,0x00,0xf0,0xa8,0xa8,0xa8,0xa8,0x00,
		0x00,0x00,0xb0,0xc8,0x88,0x88,0x88,0x00,0x00,0x00,0x70,0x88,0x88,0x88,0x70,0x00,
		0x00,0x00,0xf0,0x88,0x88,0xf0,0x80,0x80,0x00,0x00,0x78,0x88,0x88,0x78,0x08,0x08,
		0x00,0x00,0xb0,0xc8,0x80,0x80,0x80,0x00,0x00,0x00,0x78,0x80,0x70,0x08,0xf0,0x00,
		0x20,0x20,0x70,0x20,0x20,0x20,0x18,0x00,0x00,0x00,0x88,0x88,0x88,0x98,0x68,0x00,
		0x00,0x00,0x88,0x88,0x88,0x50,0x20,0x00,0x00,0x00,0xa8,0xa8,0xa8,0xa8,0x50,0x00,
		0x00,0x00,0x88,0x50,0x20,0x50,0x88,0x00,0x00,0x00,0x88,0x88,0x88,0x78,0x08,0x70,
		0x00,0x00,0xf8,0x10,0x20,0x40,0xf8,0x00,0x08,0x10,0x10,0x20,0x10,0x10,0x08,0x00,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x40,0x20,0x20,0x10,0x20,0x20,0x40,0x00,
		0x00,0x68,0xb0,0x00,0x00,0x00,0x00,0x00,0x20,0x50,0x20,0x50,0xa8,0x50,0x00,0x00,
};

#define DRAW_COLOR_BLACK draw_color(0,0,0)
#define DRAW_COLOR_GRAY draw_color(196,196,196)
#define DRAW_COLOR_WHITE draw_color(255,255,255)
#define DRAW_COLOR_RED draw_color(255,0,0)
#define DRAW_COLOR_GREEN draw_color(0,255,0)
#define DRAW_COLOR_BLUE draw_color(0,0,255)

static __inline__ unsigned draw_color(unsigned r, unsigned g, unsigned b) {
	if (video_is_graphics()) {
		video_rgb c;
		video_rgb_make(&c,r,g,b);
		return c;
	} else if (video_is_text()) {
		return (r >> 7) | ((g >> 7) << 1) | ((b >> 7) << 2);
	}

	return 0;
}

static void draw_char(int x, int y, char c, unsigned color) {
	if (video_is_graphics()) {
		int i,j;
		unsigned char* p = draw_font + 8*(unsigned)c;
		x *= FONTX;
		y *= FONTY;
			for(i=0;i<FONTY;++i) {
			for(j=0;j<FONTX;++j) {
				if (p[i] & (1 << (7-j)))
					video_put_pixel(x+j,y+i,color);
				else
					video_put_pixel(x+j,y+i,DRAW_COLOR_BLACK);
			}
		}
	} else if (video_is_text()) {
		video_put_char(x,y,c,color);
	}
}

static void draw_string(int x, int y, const char* s, unsigned color) {
	unsigned len = strlen(s);
	while (len) {
		draw_char(x,y,*s, color);
		++s;
		--len;
		++x;
	}
}

static void draw_text_fill(int x, int y, char c, int dx, unsigned color) {
	while (dx>0) {
		screen_put(x,y,c,color);
		--dx;
		++x;
	}
}

static void draw_text_fillrect(int x, int y, char c, int dx, int dy, unsigned color) {
	while (dy-->0) {
		draw_text_fill(x,y++,c,dx,color);
	}
}

static void draw_text_left(int x, int y, int dx, const char* s, unsigned color) {
	int len = strlen(s);
	int i;
	for(i=0;i<dx;++i) {
		if (i<len)
			screen_put(x+i,y,s[i], color);
		else
			screen_put(x+i,y,' ', color);
	}
}

static void draw_text_center(int x, int y, int dx, const char* s, unsigned color) {
	int len = strlen(s);
	int pre;
	int post;
	if (len > dx)
		len = dx;

	pre = (dx - len + 1) / 2;
	post = dx - pre - len;

	draw_text_fill(x,y,' ',pre,color);
	draw_text_left(x+pre,y,len,s,color);
	draw_text_fill(x+pre+len,y,' ',post,color);
}

static unsigned draw_text_string(int x, int y, const char* s, unsigned color) {
	int len = strlen(s);
	if (x + len > screen_size_x())
		len = screen_size_x() - x;
	draw_text_left(x,y,len,s,color);
	return len;
}

static int draw_text_para(int x, int y, int dx, int dy, const char* s, unsigned color) {
	int iy = 0;
	int ix = 0;
	int needspace = 0;
	while (*s && iy<=dy) {
		unsigned l = strcspn(s," \n");
		if (needspace) {
		if (ix<dx) {
				draw_text_fill(x+ix,y+iy,' ',1,color);
				++ix;
			} else {
				++iy;
				ix = 0;
			}
			needspace = 0;
		}
		if (ix+l>dx) {
			draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
			++iy;
			ix = 0;
			needspace = 0;
		}
		if (iy<=dy) {
			draw_text_left(x+ix,y+iy,l,s,color);
			ix += l;
			s += l;
			if (*s==' ') {
				++s;
				needspace = 1;
			} else if (*s=='\n') {
				draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
				++s;
				++iy;
				ix = 0;
				needspace = 0;
			}
		}
	}
	if (iy<=dy && ix<dx) {
		draw_text_fill(x+ix,y+iy,' ',dx-ix,color);
		++iy;
	}
	return y+iy;
}

static int draw_text_read(int x, int y, char* s, int dx, unsigned color) {
	int maxlen = dx - 1;
	if (strlen(s) > maxlen) {
		s[maxlen] = 0;
	}

	while (1) {
		int userkey;
		int len = strlen(s);
		draw_text_string(x,y,s,color);
		draw_text_fill(x+len,y,' ',dx-len,color);
		userkey = input_getkey();
		switch (userkey) {
			case INPUT_KEY_BACKSPACE :
			case INPUT_KEY_DEL :
				if (len) {
					s[len-1] = 0;
				}
				break;
			case INPUT_KEY_ENTER :
			case INPUT_KEY_ESC :
				return userkey;
			default:
				if (len < maxlen) {
					if (userkey>=32 && userkey<=126) {
						s[len] = userkey;
						s[len+1] = 0;
					}
				}
		}
	}
}

/* Set the default color palette */
static void draw_graphics_palette(void)
{
	if (video_is_graphics()) {
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
			video_index_packed_to_rgb(0);
		}
	}
}

/* Draw a test screen */
static void draw_default_test(void) {
	int i,j;

	int TEST_DELTA1 = 10;
	int TEST_DELTA2 = 2;

	draw_graphics_palette();

	if (video_is_graphics()) {
		video_rgb c;

		/* draw out of screen data */
		video_rgb_make(&c,255,0,0);
		for(i=0;i<video_size_x();++i)
			video_put_pixel(i,video_size_y(),c);
		for(j=0;j<video_size_y();++j)
			video_put_pixel(video_size_x(),j,c);

		/* draw background */
		video_rgb_make(&c,0,0,0);
		video_clear(0,0,video_size_x(),video_size_y(),c);

		/* draw border */
		video_rgb_make(&c,255,255,255);
		for(i=0;i<video_size_x();++i) {
			video_put_pixel(i,0,c);
			video_put_pixel(i,video_size_y()-1,c);
		}
		for(j=0;j<video_size_y();++j) {
			video_put_pixel(0,j,c);
			video_put_pixel(video_size_x()-1,j,c);
		}

		/* draw grid */
		video_rgb_make(&c,128,128,128);
		for(i=TEST_DELTA1;i<video_size_x()-1;i += TEST_DELTA1)
			for(j=TEST_DELTA2;j<video_size_y()-1;j += TEST_DELTA2)
				video_put_pixel(i,j,c);
		for(j=TEST_DELTA1;j<video_size_y()-1;j += TEST_DELTA1)
			for(i=TEST_DELTA2;i<video_size_x()-1;i += TEST_DELTA2)
				video_put_pixel(i,j,c);

		/* draw line */
		video_rgb_make(&c,196,196,196);
		for(i=1;i<video_size_x()/2 && i<video_size_y()/2;++i) {
			video_put_pixel(i,i,c);
			video_put_pixel(video_size_x()-i-1,i,c);
			video_put_pixel(i,video_size_y()-i-1,c);
			video_put_pixel(video_size_x()-i-1,video_size_y()-i-1,c);
		}

	} else if (video_is_text()) {
		/* draw out of screen data */
		for(i=0;i<screen_size_x();++i) {
			video_put_char(i,screen_size_y(),'',DRAW_COLOR_RED);
		}
		for(j=0;j<screen_size_y();++j) {
			video_put_char(screen_size_x(),j,'',DRAW_COLOR_RED);
		}

		/* draw border */
		for(i=0;i<screen_size_x();++i) {
			video_put_char(i,0,'',DRAW_COLOR_GRAY);
			video_put_char(i,screen_size_y()-1,'',DRAW_COLOR_GRAY);
		}
		for(j=0;j<screen_size_y();++j) {
			video_put_char(0,j,'',DRAW_COLOR_GRAY);
			video_put_char(screen_size_x()-1,j,'',DRAW_COLOR_GRAY);
		}

		/* draw grid */
		for(i=1;i+1<screen_size_x();++i) {
			for(j=1;j+1<screen_size_y();++j) {
				video_put_char(i,j,'',DRAW_COLOR_GRAY);
			}
		}

	}
}

static void draw_graphics_ellipse(int xc, int yc, int a0, int b0, unsigned color) {
	int x = 0;
	int y = b0;
	int a = a0;
	int b = b0;
	int as = a*a;
	int as2 = 2*as;
	int bs = b*b;
	int bs2 = 2*bs;
	int d = bs - as*b + as/4;
	int dx = 0;
	int dy = as2 * b;

	while (dx<dy) {
		video_put_pixel(xc+x,yc+y,color);
		video_put_pixel(xc-x,yc+y,color);
		video_put_pixel(xc+x,yc-y,color);
		video_put_pixel(xc-x,yc-y,color);
		if (d>0) {
			--y;
			dy -= as2;
			d -= dy;
		}
		++x;
		dx += bs2;
		d += bs + dx;
	}

	d += (3*(as-bs)/2 - (dx+dy)) / 2;

	while (y>=0) {
		video_put_pixel(xc+x,yc+y,color);
		video_put_pixel(xc-x,yc+y,color);
		video_put_pixel(xc+x,yc-y,color);
		video_put_pixel(xc-x,yc-y,color);
		if (d<0) {
			++x;
			dx += bs2;
			d += dx;
		}
		--y;
		dy -= as2;
		d += as - dy;
	}
}

/* Draw the animation screen */
static void draw_graphics_animate(int s_x, int s_y, int s_dx, int s_dy, unsigned counter, int do_clear) {
	if (video_is_graphics()) {
		video_rgb c;

		int dx = s_dx / (4*4);
		int dy = s_dy / (3*4);

		int x = counter % ((s_dx - dx) * 2);
		int y = counter % ((s_dy - dy) * 2);

		if (x >= (s_dx - dx))
			x = (s_dx - dx) * 2 - x;
		if (y >= (s_dy - dy))
			y = (s_dy - dy) * 2 - y;

		assert( x + dx <= s_dx );
		assert( y + dy <= s_dy );

		/* draw bar */
		if (do_clear)
			video_rgb_make(&c,0,0,0);
		else
			video_rgb_make(&c,255,255,255);
		video_clear(s_x + x,s_y,dx,s_dy,c);
		video_clear(s_x,s_y+y,s_dx,dy,c);
	}
}

/* Draw the out of screen data */
static void draw_graphics_out_of_screen(int do_clear) {
	if (video_is_graphics()) {
		video_rgb c;
		unsigned x,y;

		/* draw out of screen data */
		if (do_clear)
			video_rgb_make(&c,0,0,0);
		else
			video_rgb_make(&c,255,0,0);

		if (video_virtual_x() > video_size_x())
			x = video_size_x() + 1;
		else
			x = video_size_x();
		if (video_virtual_y() > video_size_y())
			y = video_size_y() + 1;
		else
			y = video_size_y();

		video_clear(0,0,x,y,c);
	}
}

/* Draw a test calibration screen */
static void draw_graphics_calib(int s_x, int s_y, int s_dx, int s_dy) {
	int i,j,k,l;

	draw_graphics_palette();

	if (video_is_graphics()) {
		video_rgb c;

		unsigned dx = s_dx*3/5;
		unsigned dy = s_dy*4/5;
		unsigned x = (s_dx-dx)/2;
		unsigned y = (s_dy-dy)/2;
		unsigned dyB = dy/3;
		unsigned dyM = dyB/2;
		unsigned dyS = dyB/4;

		/* draw background */
		video_rgb_make(&c,128,128,128);
		video_clear(s_x,s_y,s_dx,s_dy,c);

		/* draw rows */
		video_rgb_make(&c,255,255,255);
		for(i=0;i<15;++i)
			video_clear(s_x,s_y + s_dy*i/15+s_dy/30,s_dx,1,c);

		/* draw columns */
		for(i=0;i<20;++i) {
			unsigned rx = s_dx*i/20+s_dx/40;
			for(j=0;j<video_size_y();++j)
				video_put_pixel(s_x+rx,s_y+j,c);
		}

		/* draw ramp */
		for(i=0;i<dx;++i) {
			switch (i*8/dx) {
				case 0 : video_rgb_make(&c,255,255,255); break;
				case 1 : video_rgb_make(&c,255,255,0); break;
				case 2 : video_rgb_make(&c,0,255,255); break;
				case 3 : video_rgb_make(&c,0,255,0); break;
				case 4 : video_rgb_make(&c,255,0,255); break;
				case 5 : video_rgb_make(&c,255,0,0); break;
				case 6 : video_rgb_make(&c,0,0,255); break;
				case 7 : video_rgb_make(&c,0,0,0); break;
			}
			for(j=0;j<dyB;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyB;

		for(i=0;i<dx;++i) {
			video_rgb_make(&c,i*255/dx,i*255/dx,i*255/dx);
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;


		k = 0;
		l = 0;
		for(i=0;i<dx;++i) {
			if (k > i / (dx / 5)) {
				l = !l;
				k = 0;
			}
			++k;
			if  (l)
				video_rgb_make(&c,0,0,0);
			else
				video_rgb_make(&c,255,255,255);
			for(j=0;j<dyM;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyM;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,i*255/(dx/2),0,0);
			else
				video_rgb_make(&c,255,(i-dx/2)*255/(dx/2),(i-dx/2)*255/(dx/2));
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,0,i*255/(dx/2),0);
			else
				video_rgb_make(&c,(i-dx/2)*255/(dx/2),255,(i-dx/2)*255/(dx/2));
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			if (i<dx/2)
				video_rgb_make(&c,0,0,i*255/(dx/2));
			else
				video_rgb_make(&c,(i-dx/2)*255/(dx/2),(i-dx/2)*255/(dx/2),255);
			for(j=0;j<dyS;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyS;

		for(i=0;i<dx;++i) {
			switch (i*5/dx) {
				case 0 : video_rgb_make(&c,0,0,0); break;
				case 1 : video_rgb_make(&c,64,64,64); break;
				case 2 : video_rgb_make(&c,128,128,128); break;
				case 3 : video_rgb_make(&c,196,196,196); break;
				case 4 : video_rgb_make(&c,255,255,255); break;
			}
			for(j=0;j<dyM;++j)
				video_put_pixel(s_x+x+i,s_y+y+j,c);
		}
		y += dyM;


		video_rgb_make(&c,255,255,255);
		draw_graphics_ellipse(s_dx / 2,  s_dy / 2, s_dx * 3 / 9, s_dy * 4 / 9, c);
	}
}

/* Do a speed test */
static unsigned draw_graphics_speed(int s_x, int s_y, int s_dx, int s_dy) {

	draw_graphics_palette();

	if (video_is_graphics()) {
		unsigned i;
		uclock_t start,stop,end;
		unsigned size,count;
		struct video_pipeline_struct pipeline;
		uint8* data;

		size = s_dx * s_dy * video_bytes_per_pixel();

		data = malloc(size);
		for(i=0;i<size;++i)
			data[i] = i;

		video_blit_pipeline_init(&pipeline, s_dx, s_dy, s_dx*video_bytes_per_pixel(), video_bytes_per_pixel());

		/* fill the cache */
		uclock();
		video_blit_pipeline(&pipeline, s_x, s_y, data);

		count = 0;
		start = uclock();
		end = start + UCLOCKS_PER_SEC * 2;
		stop = start;
		while (stop < end) {
			video_blit_pipeline(&pipeline, s_x, s_y, data);
			++count;
			stop = uclock();
		}

		video_pipeline_done(&pipeline);
		free(data);

		size = size * count * (double)UCLOCKS_PER_SEC / (stop-start);

		return size;
	}

	return 0;
}

/***************************************************************************/
/* Menu */

#define COLOR_BAR 0x70
#define COLOR_ERROR 0x4F
#define COLOR_REVERSE 0x70
#define COLOR_NORMAL 0x03
#define COLOR_BOLD 0x0F

typedef void (*entry_print)(int x, int y, int dx, void* data, int n, int selected);

int draw_text_menu(int x, int y, int dx, int dy, void* data, int mac, entry_print print) {
	int done = 0;
	int base = 0;
	int pos = 0;
	int key;

	while (!done) {
		int i;

		for(i=0;i<dy;++i) {
			if (base+i >= mac)
				draw_text_fill(x,y+i,' ',dx,COLOR_NORMAL);
			else
				print(x,y+i,dx,data,base+i,pos == i);
		}

		key = input_getkey();

		switch (key) {
			case INPUT_KEY_ENTER:
                        case INPUT_KEY_ESC:
				done = 1;
				break;
	  		case INPUT_KEY_DOWN :
				if (base + pos + 1 < mac) {
					if (pos + 1 < dy)
						++pos;
					else
						++base;
				}
	  			break;
	  		case INPUT_KEY_PGDN :
				for(i=0;i<dy;++i)
				if (base + pos + 1 < mac) {
					if (pos + 1 < dy)
						++pos;
					else
						++base;
				}
	  			break;
	  		case INPUT_KEY_UP :
				if (base + pos > 0) {
					if (pos > 0)
						--pos;
					else
						--base;
				}
	  			break;
	  		case INPUT_KEY_PGUP :
				for(i=0;i<dy;++i)
				if (base + pos > 0) {
					if (pos > 0)
						--pos;
					else
						--base;
				}
	  			break;
		}
	}

	return key == INPUT_KEY_ENTER ? base + pos : -1;
}

