/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "screen.h"
#include "generate.h"
#include "scrblit.h"
#include "vbe.h"
#include "vbeline.h"
#include "vga.h"
#include "vgaline.h"
#include "vgatweak.h"
#include "font.h"
#include "allegro.h"
#include "option.h"

#include <sys/movedata.h>
#include <dos.h>
#include <math.h>
#include <conio.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <pc.h>

/***************************************************************************/
/* Colors */

#define COLOR_TITLE 0x0B
#define COLOR_INPUT 0x0F
#define COLOR_BAD 0x0C
#define COLOR_MARK 0x30
#define COLOR_MARK_BAD 0x34
#define COLOR_SELECTED 0x70
#define COLOR_SELECTED_MARK 0x7F

#define COLOR_INFO_NORMAL 0x08
#define COLOR_INFO_TITLE 0x07

#include "common.c"

/***************************************************************************/
/* Common variable */

int menu_base;
int menu_rel;
int menu_rel_max;
int menu_base_max;
int menu_max;

/* Advance VGA/VBE */
enum advance_t {
	advance_mame, advance_menu, advance_vbe, advance_vga
} advance;

int is_advance_vbe_active;
int is_advance_vga_active;

int vm_all_modified; /* container need to be saved */
int vm_enable_load_save; /* load/save is enabled */


/***************************************************************************/
/* Common test screen */

static int draw_default_command(int x, int y) {
	draw_string(x,y,"ENTER  Save & Exit",DRAW_COLOR_WHITE);
	++y;
	draw_string(x,y,"ESC    Exit",DRAW_COLOR_WHITE);
	++y;

	return y;
}

static int draw_named_test(int x, int y) {
	char buffer[256];

	draw_default_test();

	/* draw info */
	if (video_is_graphics()) {
		sprintf(buffer,"%s, graph %dx%d [%dx%d] %d bits",video_name(),video_size_x(), video_size_y(), video_virtual_x(), video_virtual_y(), video_bits_per_pixel());
	} else if (video_is_text()) {
		sprintf(buffer,"%s, text %dx%d [%dx%d]",video_name(),video_size_x(), video_size_y(), screen_size_x(), screen_size_y());
	} else {
		*buffer = 0;
	}

	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	return y;
}

static int draw_vgatweak_test(int x, int y, video_mode* vm) {
	char buffer[256];
	struct tweak_crtc info;
	unsigned memlimit;
	double horzclock;
	double vertclock;

	tweak_crtc_get(&info,tweak_register_get,&vm->driver.vgatweak);

	horzclock = (double)info.horz_clock / 1E3;
	vertclock = (double)info.vert_clock;

	y = draw_named_test(x,y);

	if (info.is_linear)
		memlimit = 0x10000;
	else
		memlimit = 0x10000*4;

	if (info.size_x * info.size_y > memlimit && info.is_linear) {
		sprintf(buffer,"Memory size %d Kbytes too big",info.size_x * info.size_y / 1024);
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	++y;

	sprintf(buffer,"Horz  Vert");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (horzclock!=0 || vertclock!=0) {
		sprintf(buffer,"%4.1f %5.1f      Clock [kHz Hz]", horzclock, vertclock);
		draw_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	sprintf(buffer,"%4d  %4d [qa] Display End",info.hde,info.vde);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4d  %4d [ws] Blank Start",info.hbs,info.vbs);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hde<=info.hbs)) {
		sprintf(buffer,"HDE<=HBS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vde<=info.vbs)) {
		sprintf(buffer,"VDE<=VBS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [ed] Retrace Start",info.hrs,info.vrs);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hde<=info.hrs)) {
		sprintf(buffer,"HDE<=HRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vde<=info.vrs)) {
		sprintf(buffer,"VDE<=VRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.hbs<=info.hrs)) {
		sprintf(buffer,"HBS<=HRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vbs<=info.vrs)) {
		sprintf(buffer,"VBS<=VRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [rf] Retrace End",info.hre,info.vre);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hrs<info.hre)) {
		sprintf(buffer,"HRS<HRE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
			++y;
	}
	if (!(info.vrs<info.vre)) {
		sprintf(buffer,"VRS<VRE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [tg] Blank End",info.hbe,info.vbe);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hre<=info.hbe)) {
		sprintf(buffer,"HRE<=HBE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vre<=info.vbe)) {
		sprintf(buffer,"VRE<=VBE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d [yh] Total",info.ht,info.vt);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (!(info.hre<=info.ht)) {
		sprintf(buffer,"HRE<=HT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vre<=info.vt)) {
		sprintf(buffer,"VRE<=VT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	if (!(info.hbe<=info.ht)) {
		sprintf(buffer,"HBE<=HT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(info.vbe<=info.vt)) {
		sprintf(buffer,"VBE<=VT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"   %c     %c [uj] Polarization", info.hsync ? '-' : '+', info.vsync ? '-' : '+');
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4d  %4d [ik] Char Size", info.char_size_x,info.scanlines_per_line);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4d [o]  Scanline size", info.bytes_per_scanline);
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s [x]  Doublescan", info.is_doublescan ? "on" : "off");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s [c]  Middle clock", info.dotclock_middle ? "on" : "off");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"%4s       [v]  Master clock [MHz]", info.dotclock_master ? "28" : "25");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	++y;
	draw_string(x,y,"Q...I  Inc horz (SHIFT dec)",DRAW_COLOR_WHITE);
	++y;
	draw_string(x,y,"A...K  Inc vert (SHIFT dec)",DRAW_COLOR_WHITE);
	++y;
	draw_string(x,y,"XCV    Flip flag",DRAW_COLOR_WHITE);
	++y;
	draw_string(x,y,"Z      Compute blank",DRAW_COLOR_WHITE);
	++y;

	y = draw_default_command(x,y);

	return y;
}

static int draw_crtc_test(int x, int y, video_crtc* crtc, int print_clock, int print_measured_clock, int print_key) {
	char buffer[256];

	sprintf(buffer,"Horz  Vert");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (print_clock && crtc->pixelclock!=0) {
		double horzclock = (double)crtc->pixelclock / (crtc->ht * 1E3);
		double vertclock = (double)crtc->pixelclock / (crtc->ht * crtc->vt);
		if (crtc->doublescan)
			vertclock /= 2;
		if (crtc->interlaced)
			vertclock *= 2;
		sprintf(buffer,"%4.1f %5.1f %sClock Requested [kHz Hz]", horzclock, vertclock, print_key ? "     " : "");
		draw_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	if (print_measured_clock && crtc->pixelclock!=0) {
		double vertclock = video_measure_vert_freq();
		double horzclock = vertclock * crtc->vt / 1E3;
		if (crtc->doublescan)
			horzclock *= 2;
		if (crtc->interlaced)
			horzclock /= 2;
		sprintf(buffer,"%4.1f %5.1f %sClock Measured [kHz Hz]", horzclock, vertclock, print_key ? "     " : "");
		draw_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	sprintf(buffer,"%4d  %4d %sDisplay End",crtc->hde,crtc->vde, print_key ? "[qa] " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

/*
	sprintf(buffer,"%4d  %4d %sBlank Start",crtc->hbs,crtc->vbs, print_key ? "     " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
*/
/*      if (!(crtc->hde<=crtc->hbs)) {
		sprintf(buffer,"HDE<=HBS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
/*      if (!(crtc->vde<=crtc->vbs)) {
		sprintf(buffer,"VDE<=VBS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/

	sprintf(buffer,"%4d  %4d %sRetrace Start",crtc->hrs,crtc->vrs, print_key ? "[ed] " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hde<=crtc->hrs)) {
		sprintf(buffer,"HDE<=HRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vde<=crtc->vrs)) {
		sprintf(buffer,"VDE<=VRS");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"%4d  %4d %sRetrace End",crtc->hre,crtc->vre, print_key ? "[rf] " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hrs<crtc->hre)) {
		sprintf(buffer,"HRS<HRE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vrs<crtc->vre)) {
		sprintf(buffer,"VRE<VRE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

/*
	sprintf(buffer,"%4d  %4d %sBlank End",crtc->hbe,crtc->vbe, print_key ? "     " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
*/
/*
	if (!(crtc->hre<=crtc->hbe)) {
		sprintf(buffer,"HRE<=HBE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
/*
	if (!(crtc->vre<=crtc->vbe)) {
		sprintf(buffer,"VRE<=VBE");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
*/
	sprintf(buffer,"%4d  %4d %sTotal",crtc->ht,crtc->vt, print_key ? "[yh] " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;
	if (!(crtc->hre<=crtc->ht)) {
		sprintf(buffer,"HRE<=HT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}
	if (!(crtc->vre<=crtc->vt)) {
		sprintf(buffer,"VRE<=VT");
		draw_string(x,y,buffer,DRAW_COLOR_RED);
		++y;
	}

	sprintf(buffer,"   %c     %c %sPolarization", crtc->hsync ? '-' : '+', crtc->vsync ? '-' : '+', print_key ? "[uj] " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s %sDoublescan", crtc->doublescan ? "on" : "off", print_key ? "[x]  " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	sprintf(buffer,"      %4s %sInterlaced", crtc->interlaced ? "on" : "off", print_key ? "[c]  " : "");
	draw_string(x,y,buffer,DRAW_COLOR_WHITE);
	++y;

	if (print_clock) {
		sprintf(buffer,"%4.2f      %sPixelclock [MHz]", (double)crtc->pixelclock / 1E6, print_key ? "[v]  " : "");
		draw_string(x,y,buffer,DRAW_COLOR_WHITE);
		++y;
	}

	if (print_key) {
		++y;
		draw_string(x,y,"Q...U  Inc horz (SHIFT dec)",DRAW_COLOR_WHITE);
		++y;
		draw_string(x,y,"A...J  Inc vert (SHIFT dec)",DRAW_COLOR_WHITE);
		++y;
		draw_string(x,y,"XCV    Flip flag",DRAW_COLOR_WHITE);
		++y;
		draw_string(x,y,"ARROWS Center",DRAW_COLOR_WHITE);
		++y;
		draw_string(x,y,"CTRL+ARROWS Expand/Shrink",DRAW_COLOR_WHITE);
		++y;
	}

	return y;
}

static int draw_vgaline_test(int x, int y, video_mode* mode) {
	y = draw_named_test(x,y);
	++y;

	y = draw_crtc_test(x,y,&mode->driver.vgaline.crtc,1,1,1);
	y = draw_default_command(x,y);

	return y;
}

static int draw_vbeline_test(int x, int y, video_mode* mode) {
	y = draw_named_test(x,y);

	++y;

	y = draw_crtc_test(x,y,&mode->driver.vbeline.crtc,1,1,1);
	y = draw_default_command(x,y);
	return y;
}

/***************************************************************************/
/* Common information screens */

static int draw_text_warn(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	static int userkey = 0;

	if (userkey != INPUT_KEY_SPACE) {

	screen_clear();

	draw_text_center(x,y,dx,
"WARNING! Playing with video parameter can DAMAGE YOUR MONITOR!"
	,COLOR_REVERSE);
	++y;

	draw_text_center(x,y,dx,
"IF YOU HEAR STRANGE SOUNDS PRESS ESC IMMEDITIALLY!"
	,COLOR_REVERSE);
	++y;

	y = draw_text_para(x,y,dx,dy-y,
"You can modify two series of parameters, horizontal and vertical. These "
"parameters control the course of the video beam."
"\n\n"
"The display region is the part used for drawing. Start always from 0. "
"The retrace region is the part used by the beam to return backward. "
"The blank region is the part in which the beam is turned off. "
"The total end is the total number of columns and lines."
"\n\n"
"The starting and ending values of these regions are shown in the correct order, from the "
"smallest to the greatest, you must mantain this condition."
"\n\n"
"Others important values are the horizontal and vertical clock frequencies, these "
"values MUST BE COMPATIBLE with your monitor. You MUST check your monitor's manual "
"for getting correct values. "
"Generic old SVGA monitors accept fixed HClock of 31.5 and 35.5 kHz, new multisync monitors "
"accept values in the range 30-70 kHz. The generic VClock's acceptable range is 50-120 Hz."
"\n\n"
"For centering the screen modify the retrace start or use the ARROWS. "
"For changing the resolution modify the display end. "
"For resizing the screen modify the total end and the polarization or use the CTRL+ARROWS. "
"\n\n"
"Press SPACE to continue or ESC to exit."
	,COLOR_NORMAL);

	do {
		userkey = input_getkey();
	} while (userkey!=INPUT_KEY_SPACE && userkey!=INPUT_KEY_ESC);

	}

	return userkey != INPUT_KEY_SPACE;
}

static int draw_text_warn_bios(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	int userkey = 0;

	if (userkey != INPUT_KEY_Y) {

	screen_clear();

	draw_text_center(x,y,dx,
"WARNING! Using wrong video mode can DAMAGE YOUR MONITOR!"
	,COLOR_REVERSE);
	++y;

	draw_text_center(x,y,dx,
"IF YOU HEAR STRANGE SOUNDS PRESS ESC IMMEDITIALLY!"
	,COLOR_REVERSE);
	++y;

	++y;
	y = draw_text_para(x,y,dx,dy-y,
"You are going to test a BIOS VGA or VBE mode with a low frequency monitor.\n"
"Usually these modes work only with normal high frequency monitors. "
"So proceed only if you know that you are doing."
"\n\n"
"Press Y to continue or ESC to exit."
	,COLOR_NORMAL);

	do {
		userkey = input_getkey();
	} while (userkey!=INPUT_KEY_Y &&  userkey!=INPUT_KEY_ESC);

	}

	return userkey != INPUT_KEY_Y;
}

static void draw_text_warn_vbevga(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	int userkey = 0;

	screen_clear();

	if (is_advance_vbe_active) {
		draw_text_center(x,y,dx,
	"WARNING! The AdvanceVBE utility is active!"
		,COLOR_REVERSE);
	}
	if (is_advance_vga_active) {
		draw_text_center(x,y,dx,
	"WARNING! The AdvanceVGA utility is active!"
		,COLOR_REVERSE);
	}
	++y;
	++y;

	y = draw_text_para(x,y,dx,dy-y,
"Probably you want to uninstall it before run the video configuration program."
	,COLOR_NORMAL);

	if (!vm_enable_load_save) {
		++y;
		y = draw_text_para(x,y,dx,dy-y,
"To prevent problems your modelines aren't loaded and saved! You are able to test only the BIOS video modes."
		,COLOR_NORMAL);
	} else if (is_advance_vbe_active) {
		++y;
		y = draw_text_para(x,y,dx,dy-y,
"Please note that any vbe/vbeline mode created with AdvanceVBE isn't valid when AdvanceVBE isn't loaded. "
"And viceversa any vbe/vbeline mode created whitout AdvanceVBE isn't valid when AdvanceVBE is loaded.\n"
"This happen because the AdvanceVBE utility change the video mode numbers."
		,COLOR_NORMAL);
	}

	++y;
	y = draw_text_para(x,y,dx,dy-y,
"Press SPACE to continue"
	,COLOR_NORMAL);


	do {
		userkey = input_getkey();
	} while (userkey!=INPUT_KEY_SPACE);

}


static int draw_text_help(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	int userkey = 0;

	screen_clear();

	y = draw_text_para(x,y,dx,dy,
" HELP"
	,COLOR_REVERSE);

	y = draw_text_para(x,y,dx,dy-y,
"F2    Save selected modes\n"
"F3    Write to the file modes.txt a report of the selected modes\n"
"F4    Write to the file modes.c the VGA register values of the current mode\n"
"F5    Create a new video mode based on the current\n"
"F6    Duplicate the current video mode grabbing hardware parameteres\n"
"F7    Change the mode number of vgaline/vbeline modes\n"
"F8    Set an arbirary and precise pixel clock value\n"
"F9    Show a static test screen for the current video mode\n"
"F10   Show a dynamic test screen for the current video mode\n"
"SPACE Select/Unselect the current video mode\n"
"ENTER Test the current video mode\n"
"TAB   Rename the current video mode\n"
"ESC   Exit\n"
"\n"
"q/a   Increase the x/y Display End register (SHIFT for decreasing)\n"
"w/s   Increase the x/y Blank Start register (SHIFT for decreasing)\n"
"e/d   Increase the x/y Retrace Start register (SHIFT for decreasing)\n"
"r/f   Increase the x/y Retrace End register (SHIFT for decreasing)\n"
"t/g   Increase the x/y Blank End register (SHIFT for decreasing)\n"
"y/h   Increase the x/y Total End register (SHIFT for decreasing)\n"
"v     Increase the pixel clock (SHIFT for decreasing)\n"
"u/j   Switch the x/y polarization\n"
"x     On/Off the doublescan mode\n"
"c     On/Off the interlace mode\n"
"\n"
"Press ESC"
	,COLOR_NORMAL);

	do {
		userkey = input_getkey();
	} while (userkey!=INPUT_KEY_ESC);

	return userkey;
}

static int draw_text_error(void) {
	int x = 0;
	int y = 0;
	int dx = screen_size_x();
	int dy = screen_size_y();
	static int userkey = 0;

	screen_clear();

	draw_text_center(x,y,dx,
"ERROR! An error occour in your last action!"
	,COLOR_REVERSE);
	++y;

	++y;
	y = draw_text_para(x,y,dx,dy,
"Your last action generated an error. Probably you have requested an "
"unsupported feature by your hardware or software."
	,COLOR_NORMAL);

	if (*video_error_description_get()) {
		y = draw_text_para(x,y,dx,dy-y,"\nThe video software report this error:",COLOR_NORMAL);
		y = draw_text_para(x,y,dx,dy-y,video_error_description_get(),COLOR_ERROR);
	}

	y = draw_text_para(x,y,dx,dy-y,"\nPress ESC",COLOR_NORMAL);

	do {
		userkey = input_getkey();
	} while (userkey!=INPUT_KEY_ESC);

	return userkey;
}

/***************************************************************************/
/* Menu */

static void draw_text_menu_item(int x, int y, int dx, int pos, int selected) {
	char buffer[256];
	video_mode* vm = video_mode_container_pos(&the_modes,pos);

	if (vm) {
		const char* driver;
		char tag;
		unsigned color;
		char type;
		char vfreq[8];
		char hfreq[8];
		unsigned mode;

		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGA :
				driver = "vga";
				mode = vm->driver.vga.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				driver = "vgatweak";
				mode = vm->driver.vgatweak.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				driver = "vgaline";
				mode = vm->driver.vgaline.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VBE :
				driver = "vbe";
				mode = vm->driver.vbe.mode;
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				driver = "vbeline";
				mode = vm->driver.vbeline.mode;
				break;
			default:
				driver = "unknow";
				mode = 0;
		}

		switch (vm->flags & VIDEO_FLAGS_TYPE_MASK) {
			case VIDEO_FLAGS_TYPE_TEXT : type = 'T'; break;
			case VIDEO_FLAGS_TYPE_GRAPHICS : type = 'G'; break;
			default: type = '?';
		}

		if (selected) {
			if (vm->flags & VIDEO_FLAGS_USER_BIT0) {
				tag = '';
				color = COLOR_SELECTED_MARK;
			} else {
				tag = ' ';
				color = COLOR_SELECTED;
			}
		} else {
			if (vm->flags & VIDEO_FLAGS_USER_BIT0) {
				tag = '';
				if (monitor_check(vm)) {
					color = COLOR_MARK;
				} else {
					color = COLOR_MARK_BAD;
				}
			} else {
				tag = ' ';
				if (monitor_check(vm)) {
					color = COLOR_NORMAL;
				} else {
					color = COLOR_BAD;
				}
			}
		}

		if (vm->vert_freq) {
			sprintf(vfreq,"%6.2f",(double)vm->vert_freq);
		} else {
			sprintf(vfreq,"      ");
		}

		if (vm->horz_freq) {
			sprintf(hfreq,"%6.2f",(double)vm->horz_freq / 1E3);
		} else {
			sprintf(hfreq,"      ");
		}

		sprintf(buffer," %c %8s(%4x) %c %4d %4d %2d %s %s %s",
			tag,
			driver,
			mode,
			type,
			vm->size_x,
			vm->size_y,
			vm->bits,
			vfreq,
			hfreq,
			vm->name
		);

		draw_text_left(x,y,dx,buffer,color);

	} else {
		draw_text_fill(x,y,' ',dx,COLOR_NORMAL);
	}
}

static void draw_text_run_menu(int x, int y, int dx, int dy) {
	unsigned i;
	for(i=0;i<dy;++i) {
		if (menu_base + i < menu_max) {
			draw_text_menu_item(x,y+i,dx, menu_base + i, i == menu_rel);
		} else
			draw_text_fill(x,y+i,' ',dx,COLOR_NORMAL);
	}
}

/***************************************************************************/
/* Draw information bars */

static void format_info_vgatweak(char* buffer0, char* buffer1, char* buffer2, vgatweak_video_mode* mode) {
	struct tweak_crtc info;
	unsigned xdup;
	unsigned ydup;

	tweak_crtc_get(&info,tweak_register_get,mode);

	xdup = info.pixels_per_clock;
	if (!info.is_graphics_mode) {
		ydup = 1;
	} else {
		ydup = info.scanlines_per_line;
	}
	sprintf(buffer0,"plz clock dsen blst rtst rten blen totl ch type pclock scan");
	sprintf(buffer1,"h%c %6.2f%5d%5d%5d%5d%5d%5d%3d %s %6.2f %4d",info.hsync ? '-' : '+',(double)info.horz_clock/1E3,info.hde*xdup,info.hbs*xdup,info.hrs*xdup,info.hre*xdup,info.hbe*xdup,info.ht*xdup, info.char_size_x, !info.is_graphics_mode ? "text" : "grph" ,(double)(info.pixel_clock/1E6),info.bytes_per_scanline);
	sprintf(buffer2,"v%c %6.2f%5d%5d%5d%5d%5d%5d%3d%s%s",info.vsync ? '-' : '+',(double)info.vert_clock,info.vde/ydup,info.vbs/ydup,info.vrs/ydup,info.vre/ydup,info.vbe/ydup,info.vt/ydup, info.scanlines_per_line, info.is_doublescan ? " doublescan" : "", info.vbe != info.vbe7 ? " blank8bit" : "");
}

static void format_info_crtc(char* buffer0, char* buffer1, char* buffer2, const video_crtc* crtc) {
	double HD,HF,HS,HB;
	double VD,VF,VS,VB;
	double f;

	double horzclock = (double)crtc->pixelclock / (crtc->ht * 1E3);
	double vertclock = (double)crtc->pixelclock / (crtc->ht * crtc->vt);

	if (crtc->doublescan)
		vertclock /= 2;
	if (crtc->interlaced)
		vertclock *= 2;

	HD = crtc->hde;
	HF = crtc->hrs - crtc->hde;
	HS = crtc->hre - crtc->hrs;
	HB = crtc->ht - crtc->hre;
	f = 1 / (HD+HF+HS+HB);
	HD *= f;
	HF *= f;
	HS *= f;
	HB *= f;

	VD = crtc->vde;
	VF = crtc->vrs - crtc->vde;
	VS = crtc->vre - crtc->vrs;
	VB = crtc->vt - crtc->vre;
	f = 1 / (VD+VF+VS+VB);
	VD *= f;
	VF *= f;
	VS *= f;
	VB *= f;

	sprintf(buffer0,"plz clock  dsen rtst rten totl  disp  front sync  back  pclock");
	sprintf(buffer1,"h%c %7.3f%5d%5d%5d%5d %6.3f%6.3f%6.3f%6.3f %8.4f",crtc->hsync ? '-' : '+',horzclock,crtc->hde,crtc->hrs,crtc->hre,crtc->ht,HD,HF,HS,HB,(double)crtc->pixelclock / 1E6);
	sprintf(buffer2,"v%c %7.3f%5d%5d%5d%5d %6.3f%6.3f%6.3f%6.3f%s%s",crtc->vsync ? '-' : '+',vertclock,crtc->vde,crtc->vrs,crtc->vre,crtc->vt,VD,VF,VS,VB,crtc->doublescan ? " doublescan" : "", crtc->interlaced ? " interlaced" : "");
}

static void format_info(char* buffer0, char* buffer1, char* buffer2, video_mode* vm) {
	*buffer0 = 0;
	*buffer1 = 0;
	*buffer2 = 0;

	if (vm) {
		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				format_info_vgatweak(buffer0,buffer1,buffer2,&vm->driver.vgatweak);
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				format_info_crtc(buffer0,buffer1,buffer2,&vm->driver.vgaline.crtc);
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				format_info_crtc(buffer0,buffer1,buffer2,&vm->driver.vbeline.crtc);
				break;
		}
	}
}

static void draw_text_info(int x, int y, int dx, int dy, int pos) {
	char buffer[3][256];
	video_mode* vm = video_mode_container_pos(&the_modes,pos);
	format_info(buffer[0],buffer[1],buffer[2],vm);

	draw_text_left(x,y+0,dx,buffer[0],COLOR_INFO_TITLE);
	draw_text_left(x,y+1,dx,buffer[1],COLOR_INFO_NORMAL);
	draw_text_left(x,y+2,dx,buffer[2],COLOR_INFO_NORMAL);

	if (vm) {
		if (!monitor_check(vm)) {
			draw_text_left(x+dx-14,y,14," OUT OF RANGE ", COLOR_ERROR);
		}
	} else {
		draw_text_fillrect(x,y,' ',dx,dy,COLOR_INFO_NORMAL);
	}
}

static void draw_text_bar(int x, int by1, int by2, int dx) {
	char buffer[256];
	switch (advance) {
 		case advance_vbe :
			sprintf(buffer," AdvanceVBE Video Config - " __DATE__ );
			break;
		case advance_vga :
			sprintf(buffer," AdvanceVGA Video Config - " __DATE__ );
			break;
		case advance_mame :
			sprintf(buffer," AdvanceMAME Video Config - " __DATE__ );
			break;
		case advance_menu :
			sprintf(buffer," AdvanceMENU Video Config - " __DATE__ );
			break;
	}

	draw_text_left(x,by1,dx,buffer,COLOR_BAR);

	if (vbeline_active()) {
		sprintf(buffer,"Driver `%s' ",vbeline_driver_name());
	} else {
		sprintf(buffer,"Driver `none' ");
	}
	draw_text_left(x + dx - strlen(buffer),by1,strlen(buffer),buffer,COLOR_BAR);

	sprintf(buffer,
" #   type(number) m  x    y    c vclock hclock name");
	draw_text_left(x,by1+1,dx,buffer,COLOR_TITLE);

	sprintf(buffer," F1 Help  F2 Save  SPACE Select  TAB Rename  DEL Del  ENTER Test  ESC Exit");
	draw_text_left(x,by2,dx,buffer,COLOR_BAR);
}

/***************************************************************************/
/* Execute userkey action */

static void tweak_vert_scale(unsigned _mul, unsigned _div, vgatweak_video_mode* mode, struct tweak_crtc* info) {
	tweak_vde_set(info->vde * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vbs_set(info->vbs * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vrs_set(info->vrs * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vre_set(info->vre * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vbe_set(info->vbe * _mul / _div, tweak_register_set, tweak_register_get, mode);
	tweak_vtt_set(info->vt * _mul / _div, tweak_register_set, tweak_register_get, mode);
}

static int exe_test_vgatweak(int userkey, vgatweak_video_mode* mode, struct tweak_crtc* info) {
	int modify = 0;
	switch (userkey) {
		case 'o' :
			modify = 1;
			tweak_offset_set(info->clocks_per_scanline/info->memory_mode + 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'O' :
			modify = 1;
			tweak_offset_set(info->clocks_per_scanline/info->memory_mode - 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'i' :
			modify = 1;
			tweak_char_size_x_set(info->char_size_x + 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'I' :
			modify = 1;
			tweak_char_size_x_set(info->char_size_x - 1, tweak_register_set, tweak_register_get, mode);
			break;
		case 'k' :
			modify = 1;
			tweak_vert_scale(info->scanlines_per_line+1,info->scanlines_per_line,mode,info);
			tweak_char_size_y_set(info->scanlines_per_line + 1 + info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'K' :
			modify = 1;
			tweak_vert_scale(info->scanlines_per_line-1,info->scanlines_per_line,mode,info);
			tweak_char_size_y_set(info->scanlines_per_line - 1 - info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'x' :
		case 'X' :
			modify = 1;
			if (info->is_doublescan)
				tweak_vert_scale(1,2,mode,info);
			else
				tweak_vert_scale(2,1,mode,info);
			tweak_doublescan_set(!info->is_doublescan, tweak_register_set, tweak_register_get, mode);
			break;
		case 'u' :
		case 'U' :
			modify = 1;
			tweak_hsync_set(!info->hsync, tweak_register_set, tweak_register_get, mode);
			break;
		case 'j' :
		case 'J' :
			modify = 1;
			tweak_vsync_set(!info->vsync, tweak_register_set, tweak_register_get, mode);
			break;
		case 'c' :
		case 'C' :
			modify = 1;
			tweak_dotclock_middle_set(!info->dotclock_middle, tweak_register_set, tweak_register_get, mode);
			break;
		case 'v' :
		case 'V' :
			modify = 1;
			tweak_masterclock_input_set(!info->dotclock_master, tweak_register_set, tweak_register_get, mode);
			break;
		case 'z' :
		case 'Z' :
			modify = 1;
			info->hbs = (info->hrs + info->hde) / 2;
			info->hbe = (info->hre + info->ht) / 2;
			tweak_hbs_set(info->hbs, tweak_register_set, tweak_register_get, mode);
			tweak_hbe_set(info->hbe, tweak_register_set, tweak_register_get, mode);
			info->vbs = (info->vrs + info->vde) / 2;
			info->vbe = (info->vre + info->vt) / 2;
			tweak_vbs_set(info->vbs, tweak_register_set, tweak_register_get, mode);
			tweak_vbe_set(info->vbe, tweak_register_set, tweak_register_get, mode);
			break;
		case 'q' :
			tweak_hde_set(info->hde + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'Q' :
			tweak_hde_set(info->hde - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'w' :
			tweak_hbs_set(info->hbs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'W' :
			tweak_hbs_set(info->hbs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'e' :
			tweak_hrs_set(info->hrs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'E' :
			tweak_hrs_set(info->hrs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'r' :
			tweak_hre_set(info->hre + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'R' :
			tweak_hre_set(info->hre - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 't' :
			tweak_hbe_set(info->hbe + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'T' :
			tweak_hbe_set(info->hbe - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'y' :
			tweak_htt_set(info->ht + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'Y' :
			tweak_htt_set(info->ht - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'a' :
			tweak_vde_set(info->vde + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'A' :
			tweak_vde_set(info->vde - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 's' :
			tweak_vbs_set(info->vbs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'S' :
			tweak_vbs_set(info->vbs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'd' :
			tweak_vrs_set(info->vrs + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'D' :
			tweak_vrs_set(info->vrs - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'f' :
			tweak_vre_set(info->vre + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'F' :
			tweak_vre_set(info->vre - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'g' :
			tweak_vbe_set(info->vbe + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'G' :
			tweak_vbe_set(info->vbe - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'h' :
			tweak_vtt_set(info->vt + 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
		case 'H' :
			tweak_vtt_set(info->vt - 1, tweak_register_set, tweak_register_get, mode);
			modify = 1;
			break;
	}
	return modify;
}

static int exe_test_crtc(int userkey, video_crtc* crtc, int xdelta) {
	int modify = 0;
	unsigned pred_t;
	switch (userkey) {
		case INPUT_KEY_CTRLLEFT :
			pred_t = crtc->ht;
			crtc->ht -= crtc->ht % xdelta;
			crtc->ht -= xdelta;
			crtc->pixelclock = (double)crtc->pixelclock * crtc->ht / pred_t;
			modify = 1;
			break;
		case INPUT_KEY_CTRLRIGHT :
			pred_t = crtc->ht;
			crtc->ht -= crtc->ht % xdelta;
			crtc->ht += xdelta;
			crtc->pixelclock = (double)crtc->pixelclock * crtc->ht / pred_t;
			modify = 1;
			break;
		case INPUT_KEY_CTRLUP :
			crtc->vt -= 1;
			modify = 1;
			break;
		case INPUT_KEY_CTRLDOWN :
			crtc->vt += 1;
			modify = 1;
			break;
		case INPUT_KEY_LEFT :
			crtc->hrs -= crtc->hrs % xdelta;
			crtc->hrs += xdelta;
			crtc->hre -= crtc->hre % xdelta;
			crtc->hre += xdelta;
			modify = 1;
			break;
		case INPUT_KEY_RIGHT :
			crtc->hrs -= crtc->hrs % xdelta;
			crtc->hrs -= xdelta;
			crtc->hre -= crtc->hre % xdelta;
			crtc->hre -= xdelta;
			modify = 1;
			break;
		case INPUT_KEY_DOWN :
			crtc->vrs -= 1;
			crtc->vre -= 1;
			modify = 1;
			break;
		case INPUT_KEY_UP :
			crtc->vrs += 1;
			crtc->vre += 1;
			modify = 1;
			break;
		case 'x' :
		case 'X' :
			modify = 1;
			crtc->doublescan = !crtc->doublescan;
			break;
		case 'u' :
		case 'U' :
			modify = 1;
			crtc->hsync = !crtc->hsync;
			break;
		case 'j' :
		case 'J' :
			modify = 1;
			crtc->vsync = !crtc->vsync;
			break;
		case 'e' :
			crtc->hrs -= crtc->hrs % xdelta;
			crtc->hrs += xdelta;
			modify = 1;
			break;
		case 'E' :
			crtc->hrs -= crtc->hrs % xdelta;
			crtc->hrs -= xdelta;
			modify = 1;
			break;
		case 'r' :
			crtc->hre -= crtc->hre % xdelta;
			crtc->hre += xdelta;
			modify = 1;
			break;
		case 'R' :
			crtc->hre -= crtc->hre % xdelta;
			crtc->hre -= xdelta;
			modify = 1;
			break;
		case 'y' :
			crtc->ht -= crtc->ht % xdelta;
			crtc->ht += xdelta;
			modify = 1;
			break;
		case 'Y' :
			crtc->ht -= crtc->ht % xdelta;
			crtc->ht -= xdelta;
			modify = 1;
			break;
		case 'q' :
			crtc->hde -= crtc->hde % xdelta;
			crtc->hde += xdelta;
			modify = 1;
			break;
		case 'Q' :
			crtc->hde -= crtc->hde % xdelta;
			crtc->hde -= xdelta;
			modify = 1;
			break;
		case 'a' :
			++crtc->vde;
			modify = 1;
			break;
		case 'A' :
			--crtc->vde;
			modify = 1;
			break;
		case 'd' :
			++crtc->vrs;
			modify = 1;
			break;
		case 'D' :
			--crtc->vrs;
			modify = 1;
			break;
		case 'f' :
			++crtc->vre;
			modify = 1;
			break;
		case 'F' :
			--crtc->vre;
			modify = 1;
			break;
		case 'h' :
			++crtc->vt;
			modify = 1;
			break;
		case 'H' :
			--crtc->vt;
			modify = 1;
			break;
	}
	return modify;
}

static int exe_test_vgaline(int userkey, vgaline_video_mode* mode) {
	video_crtc* crtc = &mode->crtc;
	unsigned oldpixelclock = crtc->pixelclock;
	unsigned xdelta;
	int modify;
	if (mode->mode < 0x4) {
		if (mode->crtc.hde < 360)
			xdelta = 8;
		else if (mode->crtc.hde < 500)
			xdelta = 9;
		else if (mode->crtc.hde < 720)
			xdelta = 8;
		else
			xdelta = 9;
	} else
		xdelta = 4;
	modify = exe_test_crtc(userkey,crtc, xdelta);
	switch (userkey) {
		case 'v' :
			vga_pixelclock_getnext(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
		case 'V' :
			vga_pixelclock_getpred(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
	}

	return modify;
}

static int exe_test_vbeline(int userkey, vbeline_video_mode* mode) {
	video_crtc* crtc = &mode->crtc;
	unsigned oldpixelclock = crtc->pixelclock;
	int modify = exe_test_crtc(userkey,crtc,8);
	switch (userkey) {
		case 'c' :
		case 'C' :
			modify = 1;
			crtc->interlaced = !crtc->interlaced;
			break;
		case 'v' :
			vbeline_pixelclock_getnext(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
		case 'V' :
			vbeline_pixelclock_getpred(&crtc->pixelclock,mode->mode);
			modify = oldpixelclock != crtc->pixelclock;
			break;
	}
	return modify;
}

static void exe_vsync(video_mode* vm) {
	uclock_t start,stop;

	/* inizialize uclock (needed to reprogram the PIC) */
	uclock();

	video_wait_vsync();
	start = uclock();
	video_wait_vsync();
	stop = uclock();

	if (stop!=start)
		vm->vert_freq = (double)UCLOCKS_PER_SEC/(stop-start);
	else
		vm->vert_freq = 0;

	/* check valid value */
	if (vm->vert_freq<25 || vm->vert_freq>200)
		vm->vert_freq = 0;
}

/***************************************************************************/
/* Menu command */

static void cmd_select(int i) {
	video_mode* vm;

	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&the_modes,i);
	if (vm->flags & VIDEO_FLAGS_USER_BIT0)
		vm->flags &= ~VIDEO_FLAGS_USER_BIT0;
	else
		vm->flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
}

static int cmd_offvideo_test(int i, int userkey) {
	video_mode* vm;
	struct tweak_crtc info;
	video_mode vm_save;
	int vm_save_modified;
	int modify = 0;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);

	vm_save = *vm;
	vm_save_modified = vm_all_modified;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			tweak_crtc_get(&info, tweak_register_get, &vm->driver.vgatweak);
			modify = exe_test_vgatweak(userkey,&vm->driver.vgatweak,&info);
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			modify = exe_test_vgaline(userkey,&vm->driver.vgaline);
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			modify = exe_test_vbeline(userkey,&vm->driver.vbeline);
			break;
		default:
			modify = 0;
	}

	vm_all_modified =  vm_all_modified || modify;

	if (modify) {
		if (video_mode_import(vm)!=0) {
			/* abort */
			*vm = vm_save;
			vm_all_modified = vm_save_modified;
			return -1;
		}

		if (strcmp(vm->name,DEFAULT_TEXT_MODE)==0) {
			if (!monitor_check(vm) && monitor_check(&vm_save)) {
				*vm = vm_save;
				vm_all_modified = vm_save_modified;

				sound_error();

				return 0;
			}
			screen_reset();
		}
	} else {
		sound_warn();
	}

	return 0;
}

static int cmd_onvideo_test(int i) {
	video_mode* vm;
	int done ;
	video_mode vm_save;
	int vm_save_modified;
	int dirty = 1;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return 0;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn()!=0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	screen_set_font();

	vm_save = *vm;
	vm_save_modified = vm_all_modified;

	done = 0;
	while (!done) {
		int userkey;
		int modify = 0;

		int x = 1; /* default border */
		int y = 1;
		struct tweak_crtc info;
		video_crtc crtc;

		video_mode vm_last = *vm;
		int vm_last_modified = vm_all_modified;

		if (dirty) {
			double vert_freq;
			switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
				case VIDEO_FLAGS_DRIVER_VGATWEAK :
					/* import specific vgatweak information */
					draw_vgatweak_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VGALINE :
					draw_vgaline_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VBELINE :
					draw_vbeline_test(x,y,vm);
					break;
				case VIDEO_FLAGS_DRIVER_VGA :
				case VIDEO_FLAGS_DRIVER_VBE :
					y = draw_named_test(x,y);
					tweak_crtc_get(&info,tweak_reg_read,0);
					vert_freq = video_measure_vert_freq();
					if (video_crtc_import(&crtc,&info,vm,vert_freq)==0) {
						++y;
						y = draw_crtc_test(x,y,&crtc,0,1,0);
					}
					++y;
					y = draw_default_command(x,y);
					break;
			}
			dirty = 0;
		}

		userkey = input_getkey();
		switch (userkey) {
			case INPUT_KEY_ESC :
				done = 1;
				/* restore */
				*vm = vm_save;
				vm_all_modified = vm_save_modified;
				break;
			case INPUT_KEY_ENTER :
				done = 1;
				break;
			case 'm' :
			case 'M' :
				exe_vsync(vm);
				break;
		}
		switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
			case VIDEO_FLAGS_DRIVER_VGATWEAK :
				modify = exe_test_vgatweak(userkey,&vm->driver.vgatweak,&info);
				break;
			case VIDEO_FLAGS_DRIVER_VGALINE :
				modify = exe_test_vgaline(userkey,&vm->driver.vgaline);
				break;
			case VIDEO_FLAGS_DRIVER_VBELINE :
				modify = exe_test_vbeline(userkey,&vm->driver.vbeline);
				break;
			default:
				modify = 0;
		}

		vm_all_modified =  vm_all_modified || modify;

		if (!done) {
			if (modify) {
				dirty = 1;

				if (video_mode_import(vm)==0) {
					if (!monitor_check(vm) && monitor_check(&vm_last)) {
						/* restore */
						*vm = vm_last;
						vm_all_modified = vm_last_modified;

						dirty = 0;
						sound_error();
					} else {
						if (video_mode_set(vm) != 0) {
							/* abort */
							*vm = vm_save;
							vm_all_modified = vm_save_modified;
							return -1;
						}
						screen_set_font();
					}
				} else {
					/* abort */
					*vm = vm_save;
					vm_all_modified = vm_save_modified;
					return -1;
				}
			} else {
				sound_warn();
			}
		}
	}

	return 0;
}

static int cmd_onvideo_calib(int i) {
	video_mode* vm;
/*
	unsigned speed;
	char buffer[128];
*/

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);

	if (!video_mode_is_graphics(vm)) {
		video_error_description_set("Calibration screen not supported in text modes");
		return -1;
	}

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return 0;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn() != 0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	draw_graphics_palette();
	draw_graphics_out_of_screen(0);
/*
	speed = draw_graphics_speed(0,0,video_size_x(),video_size_y());
*/
	draw_graphics_calib(0,0,video_size_x(),video_size_y());

/*
	sprintf(buffer," %.2f MB/s", speed / (double)(1024*1024));
	draw_string(0,0,buffer,DRAW_COLOR_WHITE);
*/

	while (input_kbhit())
		input_getkey();

	input_getkey();

	return 0;
}

static int cmd_onvideo_animate(int i) {
	video_mode* vm;
	unsigned counter = 0;

	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);

	if (!video_mode_is_graphics(vm)) {
		video_error_description_set("Calibration screen not supported in text modes");
		return -1;
	}

	if (!monitor_check(vm))
		return -1;

	if (!monitor_check_safe(vm)) {
		if (draw_text_warn_bios() != 0)
			return 0;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (draw_text_warn() != 0)
				return 0;
		break;
	}

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	video_update_init(3);

	draw_graphics_palette();
	draw_graphics_out_of_screen(1);

	while (input_kbhit())
		input_getkey();

	while (!input_kbhit()) {
		video_update_start();
		draw_graphics_animate(video_update_x(),video_update_y(),video_size_x(),video_size_y(), counter - video_update_page_max() + 1, 1);
		++counter;
		draw_graphics_animate(video_update_x(),video_update_y(),video_size_x(),video_size_y(), counter, 0);
		video_update_stop(1);
	}

	video_update_done();

	input_getkey();

	return 0;
}

static void cmd_gotopos(int i) {
	if (i >= menu_max)
		i = menu_max - 1;
	if (i<0)
		i = 0;
	if (menu_base <= i && i < menu_base + menu_rel_max) {
		menu_rel = i - menu_base;
	} else if (i<menu_base) {
		menu_base = i;
		menu_rel = 0;
	} else {
		menu_base = i - menu_rel_max + 1;
		if (menu_base<0)
			menu_base = 0;
		menu_rel = i - menu_base;
	}
}

static int cmd_input_key(const char* tag, const char* keys) {
	draw_text_fill(0,screen_size_y()-1,' ',screen_size_x(),COLOR_REVERSE);
	draw_text_string(2,screen_size_y()-1,tag,COLOR_REVERSE);
	draw_text_string(2+strlen(tag),screen_size_y()-1,keys,COLOR_INPUT);

	while (1) {
		int i;
		int k = input_getkey();
		if (k == INPUT_KEY_ESC)
			return -1;

		for(i=0;keys[i];++i)
			if (toupper(k)==toupper(keys[i]))
				return i;
	}
}

static int cmd_input_string(const char* tag, char* buffer, unsigned length) {
	draw_text_fill(0,screen_size_y()-1,' ',screen_size_x(),COLOR_REVERSE);
	draw_text_string(2,screen_size_y()-1,tag,COLOR_REVERSE);

	if (draw_text_read(2 + strlen(tag),screen_size_y()-1,buffer,length,COLOR_INPUT) == INPUT_KEY_ENTER) {
		return 0;
	}

	return -1;
}

static void cmd_rename(void) {
	video_mode* vm;
	char buffer[128];

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&the_modes,i);

	strcpy(buffer,"");
	if (cmd_input_string(" Name : ",buffer,40)!=0)
		return;

	strcpy(vm->name,buffer);
	vm_all_modified = 1;
}

static int cmd_mode_change(void) {
	video_mode* vm;
	char buffer[128];
	unsigned mode;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGALINE :
		case VIDEO_FLAGS_DRIVER_VBELINE :
		break;
		default :
			video_error_description_set("Function supported only for VBELINE and VGALINE modes");
			return -1;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" Mode number [hex] : ",buffer,8)!=0)
		return 0;

	mode = strtol(buffer,0,16);
	if (!mode) {
		video_error_description_set("Invalid mode number");
		return -1;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			vm->driver.vgatweak.mode = mode;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			vm->driver.vgaline.mode = mode;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			vm->driver.vbeline.mode = mode;
		break;
	}

	vm_all_modified = 1;

	return 0;
}

#define generate_hfreq 0
#define generate_vfreq 1
#define generate_pfreq 2

static void video_crtc_compute(video_crtc* crtc, unsigned xres, unsigned yres, double freq, int freq_type) {
	video_generate generate;

	generate_default_vga(&generate);

	crtc->interlaced = 0;
	if (yres < 400) {
		crtc->doublescan = 1;
	} else {
		crtc->doublescan = 0;
	}
	crtc->hsync = 1;
	crtc->vsync = 1;

	generate_crtc(crtc, xres, yres, &generate);

	switch (freq_type) {
		case generate_vfreq : generate_vert_clock(crtc,freq); break;
		case generate_hfreq : generate_horz_clock(crtc,freq); break;
		case generate_pfreq : generate_pixel_clock(crtc,freq); break;
	}
}

static int cmd_mode_create(void) {
	video_mode* vm;
	video_mode copy;
	char buffer[80];
	double freq = 0;
	int freq_type;
	unsigned x;
	unsigned y;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);
	copy = *vm;
	strcpy(copy.name,"computed_mode");

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
		case VIDEO_FLAGS_DRIVER_VGALINE :
			freq_type = generate_pfreq;
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
		case VIDEO_FLAGS_DRIVER_VBE :
			switch (cmd_input_key(" Set Vertical/Horizontal/Pixel clock ? ","vhp")) {
				case 0 : freq_type = generate_vfreq; break;
				case 1 : freq_type = generate_hfreq; break;
				case 2 : freq_type = generate_pfreq; break;
				default: return 0;
			}
			break;
		default:
			video_error_description_set("Function supported only for VBELINE and VGALINE modes");
			return -1;
	}

	strcpy(buffer,"");

	switch (freq_type) {
		case generate_vfreq :
			if (cmd_input_string(" Vertical clock [Hz] (example 60.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0);
			if (freq < 10 || freq > 200) {
				video_error_description_set("Invalid vertical clock value, usually in the range 10 - 200.0 [Hz]");
				return -1;
			}
			break;
		case generate_hfreq :
			if (cmd_input_string(" Horizontal clock [kHz] (example 31.5) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E3;
			if (freq < 10*1E3 || freq > 100*1E3) {
				video_error_description_set("Invalid horizontal clock value, usually in the range 15.0 - 80.0 [kHz]");
				return -1;
			}
			break;
		case generate_pfreq :
			if (cmd_input_string(" Pixel clock [MHz] (example 14.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E6;
			if (freq < 1*1E6 || freq > 300*1E6) {
				video_error_description_set("Invalid pixel clock value, usually in the range 1.0 - 300.0 [MHz]");
				return -1;
			}
			break;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" X resolution [pixel] : ",buffer,10)!=0)
		return 0;
	x = atoi(buffer);
	if (x < 64 || x > 2048) {
		video_error_description_set("Invalid x resolution value");
		return -1;
	}

	strcpy(buffer,"");
	if (cmd_input_string(" Y resolution [pixel] : ",buffer,10)!=0)
		return 0;
	y = atoi(buffer);
	if (y < 64 || y > 2048) {
		video_error_description_set("Invalid y resolution value");
		return -1;
	}

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vga.mode;
			video_crtc_compute(&copy.driver.vgaline.crtc, x, y, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VBE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbe.mode;
			video_crtc_compute(&copy.driver.vbeline.crtc, x, y, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgaline.mode;
			video_crtc_compute(&copy.driver.vgaline.crtc, x, y, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbeline.mode;
			video_crtc_compute(&copy.driver.vbeline.crtc, x, y, freq, freq_type);
			if (video_mode_import(&copy)!=0)
				return -1;
		break;
		default:
			return -1;
		break;
	}

	copy.flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
	video_mode_container_insert_sort(&the_modes,&copy,mode_cmp);

	menu_max = video_mode_container_max(&the_modes);

	return 0;
}

static int cmd_mode_clock(void) {
	video_mode* vm;
	video_mode copy;
	char buffer[80];
	double freq = 0;
	int freq_type;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return 0;

	vm = video_mode_container_pos(&the_modes,i);
	copy = *vm;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VBELINE :
			switch (cmd_input_key(" Set Vertical/Horizontal/Pixel clock ? ","vhp")) {
				case 0 : freq_type = generate_vfreq; break;
				case 1 : freq_type = generate_hfreq; break;
				case 2 : freq_type = generate_pfreq; break;
				default: return 0;
			}
			break;
		default:
			video_error_description_set("Function supported only for VBELINE modes");
			return -1;
	}

	strcpy(buffer,"");

	switch (freq_type) {
		case generate_vfreq :
			if (cmd_input_string(" Vertical clock [Hz] (example 60.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0);
			if (freq < 10 || freq > 200) {
				video_error_description_set("Invalid vertical clock value, usually in the range 10 - 200.0 [Hz]");
				return -1;
			}
			copy.driver.vbeline.crtc.pixelclock = freq * copy.driver.vbeline.crtc.ht * copy.driver.vbeline.crtc.vt;
			if (copy.driver.vbeline.crtc.doublescan)
				copy.driver.vbeline.crtc.pixelclock *= 2;
			if (copy.driver.vbeline.crtc.interlaced)
				copy.driver.vbeline.crtc.pixelclock /= 2;
			break;
		case generate_hfreq :
			if (cmd_input_string(" Horizontal clock [kHz] (example 31.5) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E3;
			if (freq < 10*1E3 || freq > 100*1E3) {
				video_error_description_set("Invalid horizontal clock value, usually in the range 15.0 - 80.0 [kHz]");
				return -1;
			}
			copy.driver.vbeline.crtc.pixelclock = freq * copy.driver.vbeline.crtc.ht;
			break;
		case generate_pfreq :
			if (cmd_input_string(" Pixel clock [MHz] (example 14.0) : ",buffer,10)!=0)
				return 0;
			freq = strtod(buffer,0) * 1E6;
			if (freq < 1*1E6 || freq > 300*1E6) {
				video_error_description_set("Invalid pixel clock value, usually in the range 1.0 - 300.0 [MHz]");
				return -1;
			}
			copy.driver.vbeline.crtc.pixelclock = freq;
			break;
	}

	if (video_mode_import(&copy)!=0)
		return -1;

	*vm = copy;

	vm_all_modified = 1;

	return 0;
}

static int cmd_export_current(const char* name, vgatweak_video_mode* regs) {
	FILE* f;
	char buffer[3][256];
	int j;

	f = fopen("modes.h","at");
	if (!f)
		return -1;

	fprintf(f,"/*\n");

	format_info_vgatweak(buffer[0],buffer[1],buffer[2], regs);

	fprintf(f," * %s\n",buffer[0]);
	fprintf(f," * %s\n",buffer[1]);
	fprintf(f," * %s\n",buffer[2]);
	fprintf(f," */\n\n");

	fprintf(f,"Register %s[] = {\n",name);
	for(j=0;j<regs->regs_num;++j) {
		fprintf(f,"\t{ 0x%x, 0x%x, 0x%x }", regs->regs[j].port, regs->regs[j].index, regs->regs[j].value);
		if (j+1==regs->regs_num)
			fprintf(f,"\n");
		else
			fprintf(f,",\n");
	}
	fprintf(f,"};\n\n");

	fclose(f);

	return 0;
}

static int cmd_export(void) {
	video_mode* vm;
	vgatweak_video_mode regs;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return -1;

	vm = video_mode_container_pos(&the_modes,i);

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			if (video_mode_set(vm)!=0) {
				return -1;
			}
			vgatweak_mode_grab(&regs);
		break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			regs = vm->driver.vgatweak;
		break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			if (vgaline_mode_realize(&regs,&vm->driver.vgaline) != 0) {
				return -1;
			}
		break;
		default:
			sound_error();
			return 0;
	}

	return cmd_export_current(video_mode_name(vm),&regs);
}

static int cmd_mode_grab(void) {
	video_mode* vm;
	video_mode copy;
	struct tweak_crtc info;
	video_crtc crtc;
	double vert_freq;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return -1;

	vm = video_mode_container_pos(&the_modes,i);

	if (!monitor_check(vm))
		return 0;

	copy = *vm;
	strcpy(copy.name,"grabbed_mode");

	if (video_mode_set(vm)!=0) {
		return -1;
	}

	tweak_crtc_get(&info,tweak_reg_read,0);
	vert_freq = video_measure_vert_freq();

	if (video_crtc_import(&crtc,&info,vm,vert_freq)!=0)
		return -1;

	switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
		case VIDEO_FLAGS_DRIVER_VGA :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vga.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgaline.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VGATWEAK :
			copy.flags = VIDEO_FLAGS_DRIVER_VGALINE;
			copy.driver.vgaline.mode = vm->driver.vgatweak.mode;
			copy.driver.vgaline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VBE :
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbe.mode;
			copy.driver.vbeline.crtc = crtc;
			break;
		case VIDEO_FLAGS_DRIVER_VBELINE:
			copy.flags = VIDEO_FLAGS_DRIVER_VBELINE;
			copy.driver.vbeline.mode = vm->driver.vbeline.mode;
			copy.driver.vbeline.crtc = crtc;
			break;
	}

	if (video_mode_import(&copy)!=0)
		return -1;

	copy.flags |= VIDEO_FLAGS_USER_BIT0;

	vm_all_modified = 1;
	video_mode_container_insert_sort(&the_modes,&copy,mode_cmp);

	menu_max = video_mode_container_max(&the_modes);

	return 0;
}

static int cmd_report(void) {
	video_mode_container_iterator i;

	double HD = 0;
	double HF = 0;
	double HS = 0;
	double HB = 0;
	double VD = 0;
	double VF = 0;
	double VS = 0;
	double VB = 0;
	double f;

	FILE* out;

	out = fopen("modes.txt","at");
	if (!out)
		return 1;

	fprintf(out,"Video modes selected\n\n");

	for(video_mode_container_iterator_begin(&i,&the_modes);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		video_mode* vm = video_mode_container_iterator_get(&i);

		if ((vm->flags & VIDEO_FLAGS_USER_BIT0) != 0
			&& video_mode_driver(vm) == VIDEO_FLAGS_DRIVER_VBELINE) {
			const video_crtc* crtc = &vm->driver.vbeline.crtc;
			char buffer[3][256];

			HD += crtc->hde;
			HF += crtc->hrs - crtc->hde;
			HS += crtc->hre - crtc->hrs;
			HB += crtc->ht - crtc->hre;
			VD += crtc->vde;
			VF += crtc->vrs - crtc->vde;
			VS += crtc->vre - crtc->vrs;
			VB += crtc->vt - crtc->vre;

			fprintf(out,"%s\n",vm->name);
			format_info_crtc(buffer[0],buffer[1],buffer[2],crtc);
			fprintf(out,"%s\n",buffer[0]);
			fprintf(out,"%s\n",buffer[1]);
			fprintf(out,"%s\n",buffer[2]);
			fprintf(out,"\n");
		}
	}

	f = 1 / (HD+HF+HS+HB);
	HD *= f;
	HF *= f;
	HS *= f;
	HB *= f;
	f = 1 / (VD+VF+VS+VB);
	VD *= f;
	VF *= f;
	VS *= f;
	VB *= f;

	fprintf(out,"Total mean: (normalized Display/Front Porch/Sync/Back Porch)\n");
	fprintf(out,"%6.3f%6.3f%6.3f%6.3f\n",HD,HF,HS,HB);
	fprintf(out,"%6.3f%6.3f%6.3f%6.3f\n",VD,VF,VS,VB);

	fprintf(out,"\n");
	fclose(out);

	return 0;
}

static int video_mode_select_by_addr(const video_mode* a, void* b) {
	return a==b;
}

static int video_mode_select_by_hclock(const video_mode* a, void* _monitor) {
	return a->horz_freq != 0 && video_monitor_hclock_check((const video_monitor*)_monitor,a->horz_freq) == 0;
}

static int video_mode_select_by_compare(const video_mode* a, void* _b) {
	const video_mode* b = (const video_mode*)_b;
	return video_mode_compare(a,b)==0;
}
static void cmd_del(void) {
	video_mode* vm;

	int i = menu_base + menu_rel;
	if (i >= menu_max)
		return;

	vm = video_mode_container_pos(&the_modes,i);

	vm_all_modified = 1;
	video_mode_container_remove(&the_modes,video_mode_select_by_addr,vm);

	menu_max = video_mode_container_max(&the_modes);
}

static void cmd_save(void) {
	video_mode_container vm_sel;
	video_mode_container_iterator i;
	video_mode_container_init(&vm_sel);
	for(video_mode_container_iterator_begin(&i,&the_modes);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if (vm->flags & VIDEO_FLAGS_USER_BIT0)
			video_mode_container_insert(&vm_sel,vm);
	}
	video_mode_container_save(&vm_sel,set_config_string);
	video_mode_container_done(&vm_sel);

	vm_all_modified = 0;
}

static int cmd_exit(void) {
	if (vm_all_modified && vm_enable_load_save) {
		int res;

		sound_warn();

		res = cmd_input_key(" Save before exiting : ","yn");
		if (res < 0)
			return 0;

		if (res == 0)
			cmd_save();
	}

	return 1;
}

/* Window coordinate */
#define MENU_X 0
#define MENU_Y 2
#define MENU_DX (screen_size_x())
#define MENU_DY (screen_size_y()-6)

#define INFO_X 0
#define INFO_Y (screen_size_y()-4)
#define INFO_DX (screen_size_x())
#define INFO_DY 3

#define BAR_X 0
#define BAR_Y1 0
#define BAR_Y2 (screen_size_y()-1)
#define BAR_DX (screen_size_x())

/* Menu */
static int cmd_run(void) {
	int done;
	int userkey;

	menu_base = 0;
	menu_rel = 0;
	menu_rel_max = MENU_DY;
	menu_max = video_mode_container_max(&the_modes);
	menu_base_max = menu_max -  menu_rel_max;
	if (menu_base_max < 0)
		menu_base_max = 0;

	done = 0;
	while (!done) {
		draw_text_bar(BAR_X,BAR_Y1,BAR_Y2,BAR_DX);
		draw_text_info(INFO_X,INFO_Y,INFO_DX,INFO_DY,menu_base + menu_rel);
		draw_text_run_menu(MENU_X,MENU_Y,MENU_DX,MENU_DY);

		userkey = input_getkey();

		switch (userkey) {
			case INPUT_KEY_UP:
				cmd_gotopos( menu_base + menu_rel - 1);
				break;
			case INPUT_KEY_DOWN:
				cmd_gotopos( menu_base + menu_rel + 1 );
				break;
			case INPUT_KEY_HOME: {
				int i = menu_base + menu_rel - 1;
				if (i<0)
					i = 0;
				while (i>0 && !(video_mode_container_pos(&the_modes,i)->flags & VIDEO_FLAGS_USER_BIT0))
					--i;
				cmd_gotopos( i );
				break;
			}
			case INPUT_KEY_END: {
				int i = menu_base + menu_rel + 1;
				if (i >= menu_max)
					i = menu_max - 1;
				while (i < menu_max - 1 && !(video_mode_container_pos(&the_modes,i)->flags & VIDEO_FLAGS_USER_BIT0))
					++i;
				cmd_gotopos( i );
				break;
			}
			case INPUT_KEY_PGDN:
				cmd_gotopos( menu_base + menu_rel + menu_rel_max );
				break;
			case INPUT_KEY_PGUP:
				cmd_gotopos( menu_base + menu_rel - menu_rel_max );
				break;
			case INPUT_KEY_F2:
				if (vm_enable_load_save) {
					cmd_save();
				}
				break;
			case INPUT_KEY_F3:
				cmd_report();
				break;
			case INPUT_KEY_ESC:
				done = cmd_exit();
				break;
			case INPUT_KEY_SPACE:
				cmd_select( menu_base + menu_rel );
				cmd_gotopos( menu_base + menu_rel + 1 );
				break;
			case INPUT_KEY_ENTER:
				if (cmd_onvideo_test( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_F9:
				if (cmd_onvideo_calib( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_F10:
				if (cmd_onvideo_animate( menu_base + menu_rel ) != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_TAB :
				cmd_rename();
				break;
			case INPUT_KEY_F5 :
				if (cmd_mode_create() !=0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_F6 :
				if (cmd_mode_grab() !=0 ) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_F7 :
				if (cmd_mode_change() !=0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_F8 :
				if (cmd_mode_clock() !=0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			case INPUT_KEY_DEL :
				cmd_del();
				cmd_gotopos( menu_base + menu_rel );
				break;
			case INPUT_KEY_F1:
				draw_text_help();
				break;
			case INPUT_KEY_F4:
				if (cmd_export() != 0) {
					screen_reset();
					draw_text_error();
				} else {
					screen_reset();
				}
				break;
			default:
				if (cmd_offvideo_test(menu_base + menu_rel, userkey) != 0) {
					draw_text_error();
				}
				break;
		}
	}
	return userkey;
}

/***************************************************************************/
/* log */

const char* video_log_file = 0;
int video_log_sync = 0;

void video_log_va(const char *text, va_list arg)
{
	if (video_log_file) {
		FILE* f = fopen(video_log_file,"at");
		if (f) {
			vfprintf(f,text,arg);
			fclose(f);
		}
		if (video_log_sync)
			sync();
	}
}

/* RUThere code for int2f */
#define VBE_RUT_ACK1 0xAD17
#define VBE_RUT_ACK2 0x17BE
#define VGA_RUT_ACK1 0xAD17
#define VGA_RUT_ACK2 0x175A

static void advance_config(const char* argv0, const char* argv1) {
	__dpmi_regs r;
	unsigned i;

	if (strstr(argv0,"vbev.exe")!=0 || (argv1 && optionmatch(argv1,"vbev")))
		advance = advance_vbe;
	else if (strstr(argv0,"vgav.exe")!=0 || (argv1 && optionmatch(argv1,"vgav")))
		advance = advance_vga;
	else if (strstr(argv0,"mmv.exe")!=0 || (argv1 && optionmatch(argv1,"mmv")))
		advance = advance_menu;
	else
		advance = advance_mame;

	for(i=0xC0;i<=0xFF;++i) {
		r.h.ah = i;
		r.h.al = 0;
		r.x.bx = 0;
		r.x.cx = 0;
		r.x.dx = 0;
		__dpmi_int(0x2f,&r);
		if (r.h.al == 0xFF && r.x.cx == VBE_RUT_ACK1 && r.x.dx == VBE_RUT_ACK2) {
			is_advance_vbe_active = 1;
			break;
		}
	}

	for(i=0xC0;i<=0xFF;++i) {
		r.h.ah = i;
		r.h.al = 0;
		r.x.bx = 0;
		r.x.cx = 0;
		r.x.dx = 0;
		__dpmi_int(0x2f,&r);
		if (r.h.al == 0xFF && r.x.cx == VGA_RUT_ACK1 && r.x.dx == VGA_RUT_ACK2) {
			is_advance_vga_active = 1;
			break;
		}
	}
}

int my_video_mode_grab(video_mode* mode) {
	if (vbe_mode_grab(&mode->driver.vbe)==0 && vbe_mode_import(mode)==0)
		return 0;
	if (vgaline_mode_grab(&mode->driver.vgaline)==0 && vgaline_mode_import(mode)==0)
		return 0;
	if (vgatweak_mode_grab(&mode->driver.vgatweak)==0 && vgatweak_mode_import(mode)==0)
		return 0;
	return -1;
}

void grab(void) {
	video_mode vm;
	char buffer[2048];

	video_init();

	if (my_video_mode_grab(&vm) == 0) {
		strcpy(vm.name,"current");
		video_mode_print(buffer,&vm);
		printf("%s\n",buffer);
	} else {
		fprintf(stderr,"error detecting the video mode\n");
	}

	video_done();

	exit(EXIT_SUCCESS);
}

int main(int argc, char* argv[]) {
	video_mode_container vm_sel;
	video_mode_container_iterator i;

	int j;
	const char* cfg = 0;
	advance_config(argv[0],argv[1]);

	switch (advance) {
		case advance_vbe : cfg = "vbe.cfg"; break;
		case advance_vga : cfg = "vga.cfg"; break;
		case advance_menu : cfg = "menu.cfg"; break;
		case advance_mame : cfg = "mame.cfg"; break;
	}

	for (j = 1;j < argc;j++)
	{
		if (optionmatch(argv[j],"c") && j+1<argc) {
			cfg = argv[++j];
		} else if (optionmatch(argv[j],"log") || optionmatch(argv[j],"logsync")) {
			video_log_sync = optionmatch(argv[j],"logsync");
			switch (advance) {
				case advance_vbe : video_log_file = "vbev.log"; break;
				case advance_vga : video_log_file = "vgav.log"; break;
				case advance_menu : video_log_file = "mmv.log"; break;
				case advance_mame : video_log_file = "mv.log"; break;
			}
		} else if (optionmatch(argv[j],"nosound")) {
			the_sound_flag = 0;
		} else if (optionmatch(argv[j],"grab")) {
			grab();
		}
	}

	if (video_log_file)
		remove(video_log_file);

	video_log("mv: %s %s\n",__DATE__,__TIME__);

	allegro_init();
	set_config_file((char*)cfg);

	if (video_options_load(get_config_string) != 0) {
		allegro_exit();
		printf("Error loading the video options from %s\n%s\n", cfg, video_error_description_get());
		printf("Please read the file `install.txt' and `mv.txt'\n");
		exit(1);
	}

	video_init();

	if (video_monitor_load(&the_monitor,get_config_string) != 0) {
		video_done();
		allegro_exit();
		printf("Error loading the clock options from %s\n%s\n", cfg, video_error_description_get());
		printf("Please read the file `install.txt' and `mv.txt'\n");
		exit(1);
	} else {
		unsigned i;
		for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
			if (the_monitor.hclock[i].low)
				video_log("mv: hclock %.3f - %.3f\n",(double)the_monitor.hclock[i].low,(double)the_monitor.hclock[i].high);
		for(i=0;i<VIDEO_MONITOR_RANGE_MAX;++i)
			if (the_monitor.vclock[i].low)
				video_log("mv: vclock %.3f - %.3f\n",(double)the_monitor.vclock[i].low,(double)the_monitor.vclock[i].high);
	}

	/* all mode */
	video_mode_container_init(&vm_sel);

	if (advance == advance_vbe) {
		video_mode_container_insert_vbe(&vm_sel);
		if (!is_advance_vbe_active)
			video_mode_container_insert_vbeline(&vm_sel);
		video_mode_container_insert_vgaline_only_text(&vm_sel);
	} else if (advance == advance_vga) {
		video_mode_container_insert_vga(&vm_sel);
		if (!is_advance_vga_active)
			video_mode_container_insert_vgaline(&vm_sel);
		else {
			video_mode_container_insert_vgaline_only_text(&vm_sel);
			video_mode_container_insert_vgaline_only_standard(&vm_sel);
		}
	} else {
		video_mode_container_insert_vga(&vm_sel);
		video_mode_container_insert_vbe(&vm_sel);
		if (!is_advance_vga_active)
			video_mode_container_insert_vgaline(&vm_sel);
		else
			video_mode_container_insert_vgaline_only_text(&vm_sel);
		if (!is_advance_vbe_active)
			video_mode_container_insert_vbeline(&vm_sel);
	}

	/* remove */
	if (video_option_default_mode_get()==VIDEO_OPTION_DEFAULT_MODE_SMART) {
		video_mode_container_remove(&vm_sel,video_mode_select_by_hclock,&the_monitor);
	}

	/* sort */
	video_mode_container_init(&the_modes);
	for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		video_mode* vm = video_mode_container_iterator_get(&i);
		video_mode_container_insert_sort(&the_modes,vm,mode_cmp);
	}
	video_mode_container_done(&vm_sel);

	vm_enable_load_save = !(advance==advance_vbe && is_advance_vbe_active)
		&& !(advance==advance_vga && is_advance_vga_active);
	if (vm_enable_load_save ) {

		/* load selected */
		video_mode_container_init(&vm_sel);
		video_mode_container_load(&vm_sel,get_config_string);

		/* union set */
		for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			video_mode* vm = video_mode_container_iterator_get(&i);
			int has = video_mode_container_has(&the_modes,vm,video_mode_compare) != 0;
			int insert = 1;
			if (!has) {
				/* if a vesa mode is not autodetected don't insert */
				switch (vm->flags & VIDEO_FLAGS_DRIVER_MASK) {
					case VIDEO_FLAGS_DRIVER_VBE :
						insert = 0;
					break;
				}
			}
			if (insert) {
				video_mode_container_remove(&the_modes,video_mode_select_by_compare,vm);
				vm->flags |= VIDEO_FLAGS_USER_BIT0;
				video_mode_container_insert_sort(&the_modes,vm,mode_cmp);
			}
		}
		video_mode_container_done(&vm_sel);
	}

	vm_all_modified = 0;

	screen_init();

	sound_signal();

	if (advance == advance_vbe) {
		if (is_advance_vbe_active)
			draw_text_warn_vbevga();
	} else if (advance == advance_vga) {
		if (is_advance_vga_active)
			draw_text_warn_vbevga();
	} else {
		if (is_advance_vbe_active)
			draw_text_warn_vbevga();
	}

	cmd_run();

	video_log("mv: exit\n");

	screen_done();

	video_mode_container_done(&the_modes);

	video_done();
	allegro_exit();

	video_log("mv: the end\n");

	if (video_log_file)
		printf("Log saved in %s\n",video_log_file);

	return 0;
}


