/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __GENERATE_H
#define __GENERATE_H

#include "screen.h"

typedef struct video_generate_struct {
	double hactive;
	double hfront;
	double hsync;
	double hback;
	double vactive;
	double vfront;
	double vsync;
	double vback;
} video_generate;

typedef struct video_generate_linear_struct {
	unsigned hsize0;
	unsigned vsize0;
	video_generate gen0;
	unsigned hsize1;
	unsigned vsize1;
	video_generate gen1;
} video_generate_linear;

void generate_crtc(video_crtc* crtc, unsigned hsize, unsigned vsize, const video_generate* generate);
void generate_linear_interpolate(video_generate* generate, unsigned hsize, unsigned vsize, const video_generate_linear* generate_linear);

void generate_pixel_clock(video_crtc* crtc, double clock);
void generate_horz_clock(video_crtc* crtc, double clock);
void generate_vert_clock(video_crtc* crtc, double clock);

video_error generate_adjust_clock(video_crtc* crtc, const video_monitor* monitor);
video_error generate_find(video_crtc* crtc, unsigned hsize, unsigned vsize, double vclock, const video_monitor* monitor, const video_generate* generate, unsigned capability);

void generate_default_vga(video_generate* generate);
void generate_default_atari_standard(video_generate* generate);
void generate_default_atari_extended(video_generate* generate);
void generate_default_atari_medium(video_generate* generate);
void generate_default_atari_vga(video_generate* generate);
void generate_default_pal(video_generate* generate);
void generate_default_ntsc(video_generate* generate);

void generate_normalize(video_generate* generate);

video_error generate_load(video_generate* generate, const char* (*get_string)(const char *section, const char *name, const char *def));
video_error generate_parse(video_generate* generate, const char* g);
video_error generate_linear_load(video_generate_linear* generate_linear, const char* (*get_string)(const char *section, const char *name, const char *def));

#endif
