/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __SCRVGA_H
#define __SCRVGA_H

#include "scrstd.h"
#include "scrtweak.h"

#include <assert.h>
#include <go32.h>
#include <pc.h>

/* VGA status */
typedef struct vga_internal_struct {
	int active; /* != 0 if vga_init() called */

	int mode_active; /* != 0 if vga_mode_set() called */
	unsigned mode; /* mode active */

	struct tweak_crtc mode_info; /* info on mode active */

	int unchained_active; /* != 0 if unchained mode active */
	unsigned unchained_plane_mask; /* unchained plane mask */
} vga_internal;

extern vga_internal vga_state;

/***************************************************************************/
/* Public */

/* Check if vga is active */
static __inline__ int vga_active(void) {
	return vga_state.active;
}

/* Check if the one mode is active */
static __inline__ int vga_mode_active(void) {
	assert( vga_active() );
	return vga_state.mode_active;
}

/* Return the active font size */
static __inline__ unsigned vga_font_size_x(void) {
	assert( vga_mode_active() );
	return vga_state.mode_info.char_size_x;
}

static __inline__ unsigned vga_font_size_y(void) {
	assert( vga_mode_active() );
	return vga_state.mode_info.char_size_y;
}

/* Return the selector for accessing in writing the video memory */
unsigned FAST vga_write_selector(void);

/* Return the offset for accessing in writing the video memory */
unsigned FAST vga_write_line(unsigned y);

static __inline__ unsigned vga_bytes_per_scanline(void) {
	assert( vga_mode_active() );
	return vga_state.mode_info.bytes_per_scanline;
}

static __inline__ unsigned vga_adjust_bytes_per_page(unsigned bytes_per_page) {
	return (bytes_per_page + 0xFF) & ~0xFF;
}

/* Check if unchained mode is active */
static __inline__ int vga_unchained_active(void) {
	assert( vga_mode_active() );
	return vga_state.unchained_active;
}

static __inline__ void vga_unchained_plane_mask_set(unsigned plane_mask) {
	assert( vga_mode_active() );
	if (plane_mask != vga_state.unchained_plane_mask) {
		vga_state.unchained_plane_mask = plane_mask;
		outportw(TWEAK_SEQ_ADDR,0x2 | plane_mask << 8);
	}
}

/* Select the unchained plane */
static __inline__ void vga_unchained_plane_set(unsigned plane) {
	vga_unchained_plane_mask_set( 1 << plane );
}

int vga_init(void);
void vga_done(void);

int vga_font_set(void* fontraw, unsigned rows);
int vga_font_set_default8(void);
int vga_font_set_default14(void);
int vga_font_set_default16(void);

int vga_mode_set(unsigned mode, const tweak_register* regs, unsigned regs_num);
int vga_mode_get(unsigned* mode);
int vga_mode_grab(unsigned* mode, tweak_register* regs, unsigned* regs_num, unsigned regs_max);
void vga_mode_done(void);

void vga_scroll(unsigned offset, int waitvsync);
void vga_scanline_set(unsigned byte_length);
void vga_chained_set(int chained);
void vga_char_size_x_set(int size_x);
void vga_char_size_y_set(int size_y);

int vga_dotclock_get(unsigned* pixelclock);
int vga_pixelclock_get(unsigned* pixelclock, unsigned mode);
int vga_pixelclock_getnext(unsigned* pixelclock, unsigned mode);
int vga_pixelclock_getpred(unsigned* pixelclock, unsigned mode);

void vga_wait_vsync(void);

void vga_palette6_set(video_color* palette, unsigned start, unsigned count, int waitvsync);
void vga_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync);

#endif
