/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "vbeline.h"
#include "vbe.h"

#include "../card/card.h"
#include "../card/vbe3.h"
#include "../card/laguna.h"
#include "../card/s6326.h"
#include "../card/savage.h"
#include "../card/tdfx.h"
#include "../card/ati.h"
#include "../card/s3.h"
#include "../card/cirrus.h"
#include "../card/matrox.h"
#include "../card/r128.h"
#include "../card/neomagic.h"
#include "../card/trident.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/***************************************************************************/
/* Interface */

vbeline_driver cards[]=
{
	{"laguna", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_LAGUNA, detectLaguna, setLaguna, resetLaguna }, /* Laguna driver */
	{"3dfx", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | /* VBELINE_CAP_INTERLACE | */ VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_3DFX, tdfx_detect, tdfx_set, tdfx_reset }, /* 3dfx driver */
	{"savage", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_SAVAGE, detectSavage, setSavage, resetSavage }, /* Savage driver */
	{"sis6326", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_SIS6326, detectSiS6326, setSiS6326, resetSiS6326 }, /* SiS6326 driver */
	{"matrox", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_MATROX, matrox_detect, matrox_set, matrox_reset }, /* Matrox driver */
	{"r128", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT | VBELINE_CAP_REQUIRES_VBE_SAVE, VIDEO_OPTION_VBELINE_DRIVER_R128, r128_detect, r128_set, r128_reset }, /* R128 driver */
	{"neomagic", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | /* VBELINE_CAP_INTERLACE | */ VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_NEOMAGIC, neomagic_detect, neomagic_set, neomagic_reset }, /* NeoMagic driver */
	{"s3", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_S3, s3_detect, s3_set, s3_reset }, /* S3 driver */
	{"trident", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_TRIDENT, trident_detect, trident_set, trident_reset }, /* Trident driver */
	{"ati", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_ATI, detectATI, setATI, resetATI } , /* ATI driver */
	{"cirrus", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT, VIDEO_OPTION_VBELINE_DRIVER_CIRRUS, cirrus_detect, cirrus_set, cirrus_reset }, /* Cirrus driver */
	{"vbe3", VBELINE_CAP_SINGLESCAN | VBELINE_CAP_DOUBLESCAN | VBELINE_CAP_INTERLACE | VBELINE_CAP_8BIT | VBELINE_CAP_15BIT | VBELINE_CAP_16BIT | VBELINE_CAP_REQUIRES_VBE3_SET, VIDEO_OPTION_VBELINE_DRIVER_VBE3, vbe3_detect, vbe3_set, vbe3_reset }, /* VBE3 driver */
	{0,0,0,0,0,0}
};

#define is_vga_mode(mode) ((mode) < 0x100)

static void vbeline_card_setup(card_crtc *cp) {
	unsigned de,bs,rs,re,be,tt;

	cp->bits = vbeline_state.vbe_info_request.BitsPerPixel;
	cp->original_width = vbeline_state.vbe_info_request.XResolution;
	cp->original_height = vbeline_state.vbe_info_request.YResolution;

	video_crtc_realize_h(&vbeline_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->HDisp = de;
	cp->HBStart = bs;
	cp->HSStart = rs;
	cp->HSEnd = re;
	cp->HBEnd = be;
	cp->HTotal = tt;

	video_crtc_realize_v(&vbeline_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->VDisp = de;
	cp->VBStart = bs;
	cp->VSStart = rs;
	cp->VSEnd = re;
	cp->VBEnd = be;
	cp->VTotal = tt;

	/* VSyncMAME driver requires effective vertical values */
	if (vbeline_state.crtc_request.doublescan) {
		cp->VDisp *= 2;
		cp->VBStart *= 2;
		cp->VSStart *= 2;
		cp->VSEnd *= 2;
		cp->VBEnd *= 2;
		cp->VTotal *= 2;
	}
	if (vbeline_state.crtc_request.interlaced) {
		cp->VDisp /= 2;
		cp->VBStart /= 2;
		cp->VSStart /= 2;
		cp->VSEnd /= 2;
		cp->VBEnd /= 2;
		cp->VTotal /= 2;
	}

	cp->width = vbeline_state.crtc_request.hde;
	cp->height = vbeline_state.crtc_request.vde;
	cp->hpolarity = vbeline_state.crtc_request.hsync;
	cp->vpolarity = vbeline_state.crtc_request.vsync;
	cp->doublescan = vbeline_state.crtc_request.doublescan;
	cp->interlace = vbeline_state.crtc_request.interlaced;
	cp->interlaceratio = 50;

	cp->dotclockHz = vbeline_state.crtc_request.pixelclock;

	/* vga mode 13h requires double horizontal value */
	if (vbeline_state.mode_request == 0x13) {
		cp->HDisp *= 2;
		cp->HBStart *= 2;
		cp->HSStart *= 2;
		cp->HSEnd *= 2;
		cp->HBEnd *= 2;
		cp->HTotal *= 2;
		cp->dotclockHz *= 2;
	}

	/* vga mode 3h requires CRTC values computed on font of width 8 */
	if (vbeline_state.mode_request == 0x3 && (cp->HDisp % 9 == 0)) {
		cp->HDisp = cp->HDisp * 8 / 9;
		cp->HBStart = cp->HBStart * 8 / 9;
		cp->HSStart = cp->HSStart * 8 / 9;
		cp->HSEnd = cp->HSEnd * 8 / 9;
		cp->HBEnd = cp->HBEnd * 8 / 9;
		cp->HTotal = cp->HTotal * 8 / 9;
	}

	video_log("card: expected vert clock: %.2f Hz\n", (double)cp->dotclockHz / (cp->VTotal * cp->HTotal) );
	video_log_modeline_b("card: mode_set", cp->dotclockHz, cp->HDisp, cp->HBStart, cp->HSStart, cp->HSEnd, cp->HBEnd, cp->HTotal,cp->VDisp, cp->VBStart, cp->VSStart, cp->VSEnd, cp->VBEnd, cp->VTotal, cp->hpolarity, cp->vpolarity, cp->doublescan, cp->interlace);
}

void card_log(const char* text,...) {
	va_list arg;
	va_start(arg,text);
	video_log_va(text,arg);
	va_end(arg);
}

/***************************************************************************/

unsigned FAST (*vbeline_write_selector)(void);

/* Return the offset for accessing in writing the video memory */
unsigned FAST (*vbeline_write_line)(unsigned y);

void vbeline_init(void) {
	unsigned i;

	assert( !vbeline_state.active );

	/* VBE (any) is required */
	if (!vbe_active()) {
		video_log("vbeline: a vbe bios is required\n");
		return;
	}

	if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_NONE)
		return;

	vbeline_state.driver = 0;
	for(i=0;cards[i].name;++i) {
		if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_AUTO
			|| video_option_vbeline_driver_get() == cards[i].type) {
			if ((cards[i].cap & VBELINE_CAP_REQUIRES_VBE3_SET)==0 || vbe_state.info.VESAVersion >= 0x300) {
				if (cards[i].detect && cards[i].detect() > 0) {
					video_log("vbeline: found driver %s\n",cards[i].name);
					vbeline_state.driver = cards + i;
					break;
				}
			}
		}
	}

	vbeline_state.active = vbeline_state.driver != 0;
}

void vbeline_done(void) {
	assert( vbeline_active() );
	assert( !vbeline_mode_active() );

	if (vbeline_state.driver->reset)
		vbeline_state.driver->reset();

	vbeline_state.active = 0;
}

static int vbe_mode_preset(const card_crtc* ccp, vbe_ModeInfoBlock* info, vbe_CRTCInfoBlock* crtc) {
	if ((info->ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that the selected mode isn't avaliable");
		return -1;
	}

	if (ccp->interlace && (info->ModeAttributes & vbeMdInterlace)==0) {
		video_error_description_set("VBE report that the selected mode doesn't support interlace");
		return -1;
	}

	if (ccp->doublescan && (info->ModeAttributes & vbeMdDoubleScan)==0) {
		video_error_description_set("VBE report that the selected mode doesn't support doublescan");
		return -1;
	}

	if (ccp->interlace && ccp->doublescan) {
		video_error_description_set("VBE doesn't support interlaced and doublescan at the same time");
		return -1;
	}

	memset(crtc,0,sizeof(vbe_CRTCInfoBlock));

	/* if the resulting mode is smaller and if later sync values are set */
	/* use fake value to prevent a possible mode set failure if the */
	/* specified values are smaller than the fixed resolution */

	if (ccp->HDisp < info->XResolution) {
		/* use fake values, correct values are set later */
		crtc->HorizontalSyncStart = info->XResolution * 41 / 40;
		crtc->HorizontalSyncEnd = info->XResolution * 10 / 9;
		crtc->HorizontalTotal = info->XResolution * 5 / 4;
	} else {
		crtc->HorizontalTotal = ccp->HTotal;
		crtc->HorizontalSyncStart = ccp->HSStart;
		crtc->HorizontalSyncEnd = ccp->HSEnd;
	}
	if (ccp->VDisp < info->YResolution) {
		/* use fake values, correct values are set later */
		crtc->VerticalSyncStart = info->YResolution * 31 / 30;
		crtc->VerticalSyncEnd = info->YResolution * 21 / 20;
		crtc->VerticalTotal = info->YResolution * 16 / 15;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (ccp->doublescan) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	} else {
		crtc->VerticalTotal = ccp->VTotal;
		crtc->VerticalSyncStart = ccp->VSStart;
		crtc->VerticalSyncEnd = ccp->VSEnd;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (ccp->interlace) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	}

	crtc->Flags = 0;
	if (ccp->doublescan)
		crtc->Flags |= vbeCRTCDoubleScan;
	if (ccp->interlace)
		crtc->Flags |= vbeCRTCInterlaced;
	if (ccp->hpolarity)
		crtc->Flags |= vbeCRTCHorizontalSyncNegative;
	if (ccp->vpolarity)
		crtc->Flags |= vbeCRTCVerticalSyncNegative;

	crtc->PixelClock = ccp->dotclockHz;
	crtc->RefreshRate = (long)crtc->PixelClock * 100 / ((long)crtc->HorizontalTotal * crtc->VerticalTotal);

	video_log("vbeline: mode_preset %dx%d (%.2f %d %d %d %d %d %d)\n",(unsigned)info->XResolution,(unsigned)info->YResolution,(double)crtc->PixelClock/1E6,(unsigned)crtc->HorizontalSyncStart,(unsigned)crtc->HorizontalSyncEnd,(unsigned)crtc->HorizontalTotal,(unsigned)crtc->VerticalSyncStart,(unsigned)crtc->VerticalSyncEnd,(unsigned)crtc->VerticalTotal);

	return 0;
}

int vbeline_mode_set(vbeline_video_mode* mode) {
	card_crtc cp;

	assert( vbeline_active() );
	assert( !vbeline_mode_active() );

	video_log_modeline_m("vbeline: mode_set", mode->mode, mode->crtc.pixelclock, mode->crtc.hde, mode->crtc.hrs, mode->crtc.hre, mode->crtc.ht, mode->crtc.vde, mode->crtc.vrs, mode->crtc.vre, mode->crtc.vt, mode->crtc.hsync, mode->crtc.vsync, mode->crtc.doublescan, mode->crtc.interlaced );

	vbeline_state.mode_request = mode->mode;
	vbeline_state.crtc_request = mode->crtc;

	if ((vbeline_state.driver->cap & VBELINE_CAP_REQUIRES_VBE_SAVE)!=0) {
		video_log("vbeline: save video board state\n");
		if (vbe_save_state(&vbeline_state.state_size,&vbeline_state.state_ptr) != 0) {
			return -1;
		}
	}

	/* the VBE interface supports also the VGA modes */
	if (is_vga_mode(vbeline_state.mode_request)) {
		if (vga_as_vbe_mode_info_get(&vbeline_state.vbe_info_request, vbeline_state.mode_request)!=0)
			return -1;
	} else {
		if (vbe_mode_info_get(&vbeline_state.vbe_info_request, vbeline_state.mode_request)!=0)
			return -1;
	}

	if (video_option_vbeline_clock_multiplier_get()!=1.0) {
		video_log("vbeline: clock multiplier: %.2f\n",video_option_vbeline_clock_multiplier_get());
		vbeline_state.crtc_request.pixelclock *= video_option_vbeline_clock_multiplier_get();
	}

	if (vbeline_state.vbe_info_request.BitsPerPixel==8 && video_option_vbeline_8bit_clock_multiplier_get()!=1.0) {
		video_log("vbeline: 8bit_clock multiplier: %.2f\n",video_option_vbeline_8bit_clock_multiplier_get());
		vbeline_state.crtc_request.pixelclock *= video_option_vbeline_8bit_clock_multiplier_get();
	}

	if (vbeline_state.vbe_info_request.BitsPerPixel!=8 && video_option_vbeline_16bit_clock_multiplier_get()!=1.0) {
		video_log("vbeline: 16bit_clock multiplier: %.2f\n",video_option_vbeline_16bit_clock_multiplier_get());
		vbeline_state.crtc_request.pixelclock *= video_option_vbeline_16bit_clock_multiplier_get();
	}

	/* setup the card crtc info */
	vbeline_card_setup(&cp);

	if ((vbeline_state.driver->cap & VBELINE_CAP_REQUIRES_VBE3_SET)!=0) {
		if (is_vga_mode(vbeline_state.mode_request)) {
			video_error_description_set("VBE3 ins't supported for VGA modes");
			return -1;
		}
		if (vbe_pixelclock_get(&vbeline_state.crtc_request.pixelclock,vbeline_state.mode_request)!=0) {
			return -1;
		}
		if (vbe_mode_preset(&cp,&vbeline_state.vbe_info_request,&vbeline_state.vbe3_crtc_request)!=0) {
			return -1;
		}
		if (vbe_mode_set(vbeline_state.mode_request,&vbeline_state.vbe3_crtc_request)!=0) {
			return -1;
		}
	} else {
		if (is_vga_mode(vbeline_state.mode_request)) {
			if (vga_mode_set(vbeline_state.mode_request,0,0)!=0) {
				return -1;
			}
		} else {
			if (vbe_mode_set(vbeline_state.mode_request,0)!=0) {
				return -1;
			}
		}
	}

	if (is_vga_mode(vbeline_state.mode_request)) {
		if (vbeline_state.mode_request == 0x13) {
			/* convert to unchained */
			if (vbeline_state.crtc_request.hde * vbeline_state.crtc_request.vde > 0x10000)
				vga_chained_set(0);
			/* set the scanline length */
			vga_scanline_set(vbeline_state.crtc_request.hde);
		}
	} else {
		if (vbe_scanline_pixel_request(vbeline_state.crtc_request.hde) != 0) {
			vbe_mode_done();
			return -1;
		}
	}

	if (vbeline_state.driver->set(&cp) <= 0) {
		video_log("vbeline: error in the video driver mode set\n");
		vbe_mode_done();
		video_error_description_set("Error in the video driver mode set");
		return -1;
	}

	if (is_vga_mode(vbeline_state.mode_request)) {
		/* adjust resolution */
		vbeline_state.size_x = vga_state.mode_info.size_x = vbeline_state.crtc_request.hde;
		vbeline_state.size_y = vga_state.mode_info.size_y = vbeline_state.crtc_request.vde;
		/* adjust virtual size */
		vbeline_state.virtual_x = vga_state.mode_info.virtual_x;
		vbeline_state.virtual_y = vga_state.mode_info.virtual_y;
		/* write handler */
		vbeline_write_selector = vga_write_selector;
		vbeline_write_line = vga_write_line;
	} else {
		/* adjust resolution */
		vbeline_state.size_x = vbe_state.size_x = vbeline_state.crtc_request.hde;
		vbeline_state.size_y = vbe_state.size_y = vbeline_state.crtc_request.vde;
		/* adjust virtual size */
		vbeline_state.virtual_x = vbe_state.virtual_x;
		vbeline_state.virtual_y = vbe_state.virtual_y;
		/* write handler */
		vbeline_write_selector = vbe_write_selector;
		vbeline_write_line = vbe_write_line;
	}

	vbeline_state.mode_active = 1;

	return 0;
}

void vbeline_mode_done(void) {
	assert( vbeline_active() );
	assert( vbeline_mode_active() );

	if ((vbeline_state.driver->cap & VBELINE_CAP_REQUIRES_VBE_SAVE)!=0) {
		video_log("vbeline: restore video board state\n");
		vbe_restore_state(vbeline_state.state_size,vbeline_state.state_ptr);
	}

	if (vbe_mode_active())
		vbe_mode_done();
	else
		vga_mode_done();

	vbeline_state.mode_active = 0;
}

#define CARD_PIXELCLOCK_DELTA (10*1000)
#define CARD_PIXELCLOCK_MAX (250*1000*1000)
#define CARD_PIXELCLOCK_MIN (1*1000*1000)

int vbeline_pixelclock_getnext(unsigned* pixelclock, unsigned mode) {
	assert( vbeline_active() );
	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getnext(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock + CARD_PIXELCLOCK_DELTA < CARD_PIXELCLOCK_MAX)
			*pixelclock += CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

int vbeline_pixelclock_getpred(unsigned* pixelclock, unsigned mode) {
	assert( vbeline_active() );
	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getpred(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock > CARD_PIXELCLOCK_MIN - CARD_PIXELCLOCK_DELTA)
			*pixelclock -= CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

/* Import information for one video mode
	mode->driver.vbeline.mode VBE video mode number
	Return 0 if successful
*/
int vbeline_mode_import(video_mode* mode) {
	vbe_ModeInfoBlock info;

	if (!vbeline_active()) {
		video_error_description_nolog_set("VBE CRTC programming isn't avaliable on VBE version prior of 3 or without the dedicated driver");
		return -1;
	}
	if ((vbeline_state.driver->cap & VBELINE_CAP_REQUIRES_VBE3_SET)!=0 && is_vga_mode(mode->driver.vbeline.mode)) {
		video_error_description_set("VBE3 doesn't support VGA modes");
		return -1;
	}
	if (is_vga_mode(mode->driver.vbeline.mode)) {
		if (vga_as_vbe_mode_info_get(&info, mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	} else {
		if (vbe_mode_info_get(&info, mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	}
	if ((info.ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that mode %d is not avaliable", mode->driver.vbe.mode);
		return -1;
	}
	if ((info.ModeAttributes & vbeMdGraphMode) == 0) {
		if (info.MemoryModel!=vbeMemTXT) {
			video_error_description_nolog_set("Unsupported memory model");
			return -1;
		}
		// TODO
		video_error_description_nolog_set("Text modes are unsupported");
		return -1;
	} else {
		if ((mode->driver.vbeline.mode & vbeLinearBuffer) != 0) {
			if (vbe_state.info.VESAVersion < 0x200) {
				video_error_description_nolog_set("Linear frame buffer isn't available on VBE version prior 2");
				return -1;
			}
			if ((info.ModeAttributes & vbeMdLinear) == 0) {
				video_error_description_nolog_set("Linear frame buffer isn't available in mode %x", mode->driver.vbeline.mode);
				return -1;
			}
		} else {
			if ((info.ModeAttributes & vbeMdNonBanked) != 0) {
				video_error_description_nolog_set("Banked frame buffer isn't available in mode %x",mode->driver.vbeline.mode);
				return -1;
			}
		}
		/* 8,15,16 bits per pixel */
		if (info.BitsPerPixel!=8 && info.BitsPerPixel!=15 && info.BitsPerPixel!=16) {
			video_error_description_nolog_set("Unsupported bits per pixel size");
			return -1;
		}
		/* Packed or RGB memory model */
		if (info.MemoryModel!=vbeMemPK && info.MemoryModel!=vbeMemRGB) {
			video_error_description_nolog_set("Unsupported memory model");
			return -1;
		}
		/* Non planar mode */
		if (info.NumberOfPlanes!=1) {
			video_error_description_nolog_set("Unsupported number of planes");
			return -1;
		}
	}

	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		if (vbe_pixelclock_get(&mode->driver.vbeline.crtc.pixelclock,mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	} else {
		/* assume all clocks avaliable */
	}

	mode->flags = VIDEO_FLAGS_DRIVER_VBELINE |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		(mode->flags & VIDEO_FLAGS_USER_MASK);
	if ((info.ModeAttributes & vbeMdTripleBuffer) != 0)
		mode->flags |= VIDEO_FLAGS_SYNC_SETPAGE;
	if (info.MemoryModel==vbeMemTXT)
		mode->flags |= VIDEO_FLAGS_INDEX_TEXT | VIDEO_FLAGS_TYPE_TEXT;
	if (info.MemoryModel==vbeMemPK)
		mode->flags |= VIDEO_FLAGS_INDEX_PACKED | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (info.MemoryModel==vbeMemRGB)
		mode->flags |= VIDEO_FLAGS_INDEX_RGB | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (mode->driver.vbe.mode & vbeLinearBuffer)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	mode->size_x = mode->driver.vbeline.crtc.hde;
	mode->size_y = mode->driver.vbeline.crtc.vde;
	mode->virtual_x = 0;
	mode->virtual_y = 0;
	mode->vert_freq = (double)mode->driver.vbeline.crtc.pixelclock / (mode->driver.vbeline.crtc.ht * mode->driver.vbeline.crtc.vt);
	if (mode->driver.vbeline.crtc.doublescan)
		mode->vert_freq /= 2;
	if (mode->driver.vbeline.crtc.interlaced)
		mode->vert_freq *= 2;
	mode->horz_freq = (double)mode->driver.vbeline.crtc.pixelclock / mode->driver.vbeline.crtc.ht;
	mode->bits = info.BitsPerPixel;
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
	return 0;
}

/* the mode number is set later */
static const char* VBELINE_DEFAULT[] = {

/* PC MONITOR AT 35.16 kHz */
/* "pc_35.16_352x264 vbeline(0 15.81 352 369 401 449 264 274 276 284 doublescan -hsync -vsync)", */
"pc_35.16_368x276 vbeline(0 16.41 368 377 417 465 276 286 288 296 doublescan -hsync -vsync)",
"pc_35.16_384x288 vbeline(0 17.50 384 401 441 497 288 298 300 308 doublescan -hsync -vsync)",
"pc_35.16_400x300 vbeline(0 18.06 400 417 457 513 300 311 313 321 doublescan -hsync -vsync)",
"pc_35.16_432x324 vbeline(0 19.20 432 449 489 545 324 336 338 345 doublescan -hsync -vsync)",
"pc_35.16_800x600 vbeline(0 36.12 800 835 915 1027 600 614 617 642 -hsync -vsync)",

/* PC MONITOR AT 48.0 kHz */
"pc_48.0_512x384 vbeline(0 31.10 512 520 552 648 384 385 390 400 doublescan -hsync -vsync)",
"pc_48.0_1024x768 vbeline(0 61.44 1024 1040 1128 1280 768 770 780 800 -hsync -vsync)",

/* PC MONITOR AT 37.0 kHz */
"pc_37.0_1024x768 vbeline(0 47.39 1024 1040 1128 1280 768 770 780 800 -hsync -vsync interlaced)",

/* PAL SMALL resolutions */
"pal_small_368x276 vbeline(0 7.375 368 384 424 472 276 285 288 312 -hsync -vsync)",
"pal_small_736x276 vbeline(0 14.75 736 776 848 944 276 285 288 312 -hsync -vsync)",
"pal_small_736x552 vbeline(0 14.75 736 776 848 944 552 570 576 625 -hsync -vsync interlaced)",

/* PAL Special */
"pal_768x288 vbeline(0 14.75 768 792 864 944 288 291 294 312 -hsync -vsync)",

/* NTSC SMALL resolutions */
"ntsc_small_304x228 vbeline(0 6.16773 304 328 352 392 228 238 241 262 -hsync -vsync)",
"ntsc_small_608x228 vbeline(0 12.2096 608 640 696 776 228 238 241 262 -hsync -vsync)",
"ntsc_small_608x456 vbeline(0 12.2096 608 640 696 776 456 476 482 525 -hsync -vsync interlaced)",

/* NTSC Special */
"ntsc_640x242 vbeline(0 12.2096 640 664 720 776 242 245 248 262 -hsync -vsync)",

#include "vbemode.h"

0
};

static int vbeline_search_target_mode(unsigned bits, unsigned req_x, unsigned req_y) {
	video_mode best;
	int best_set = 0;

	vbe_mode_iterator vmi;
	vbe_mode_iterator_begin(&vmi);
	while (!vbe_mode_iterator_end(&vmi)) {
		video_mode v;
		int found = 0;
		memset(&v,0,sizeof(v));

		v.driver.vbe.mode = vbe_mode_iterator_get(&vmi) | vbeLinearBuffer;
		found = vbe_mode_import(&v)==0;
		if (!found) {
			v.driver.vbe.mode = vbe_mode_iterator_get(&vmi);
			found = vbe_mode_import(&v)==0;
		}

		if (found) {
			if (v.size_x >= req_x && v.size_y >= req_y && v.bits == bits) {
				if (!best_set ||
					(v.size_x <= best.size_x && v.size_y <= best.size_y)) {
					best_set = 1;
					best = v;
				}
			}
		}

		vbe_mode_iterator_next(&vmi);
	}

	if (best_set)
		return best.driver.vbe.mode;
	else
		return -1;
}

video_error vbeline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	int mode_number;
	if (mode->driver.vbeline.mode != 0) {
		if (video_mode_import(mode)==0) {
			video_mode_container_insert(vmc,mode);
		}
	} else {
		video_mode mode8 = *mode;
		video_mode mode16 = *mode;
		video_mode mode15 = *mode;

		strcat(mode8.name,"x8");
		mode_number = vbeline_search_target_mode(8,mode8.driver.vbeline.crtc.hde,mode8.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode8.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode8)==0) {
				video_mode_container_insert(vmc,&mode8);
			}
		}

		strcat(mode16.name,"x16");
		mode_number = vbeline_search_target_mode(16,mode16.driver.vbeline.crtc.hde,mode16.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode16.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode16)==0) {
				video_mode_container_insert(vmc,&mode16);
			}
		}

		strcat(mode15.name,"x15");
		mode_number = vbeline_search_target_mode(15,mode15.driver.vbeline.crtc.hde,mode15.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode15.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode15)==0) {
				video_mode_container_insert(vmc,&mode15);
			}
		}
	}

	return 0;
}

video_error vbeline_mode_generate(video_mode* mode, const char* name, const video_crtc* crtc, int bits) {
	int mode_number;

	strcpy(mode->name,name);
	mode->flags = VIDEO_FLAGS_DRIVER_VBELINE;
	mode->driver.vbeline.crtc = *crtc;
	mode->driver.vbeline.mode = 0;

	mode_number = vbeline_search_target_mode(bits,mode->driver.vbeline.crtc.hde,mode->driver.vbeline.crtc.vde);
	if (mode_number >= 0) {
		mode->driver.vbeline.mode = mode_number;
		if (video_mode_import(mode)==0) {
			return 0;
		}
	}

	return -1;
}

void video_mode_container_insert_vbeline(video_mode_container* vmc) {
	if (vbeline_active()) {
		const char** p = VBELINE_DEFAULT;
		while (*p) {
			video_mode mode;
			char* buffer = strdup(*p);
			memset(&mode,0,sizeof(mode));
			if (video_mode_parse_line(&mode,buffer)==0) {
				vbeline_mode_import_and_insert(vmc,&mode);
			}
			free(buffer);
			++p;
		}
	}
}

void video_mode_vbeline_print(char* s, const video_mode* mode) {
	s += sprintf(s,"%s vbeline(",mode->name);
	s += video_print_modeline_mc(s,mode->driver.vbeline.mode,&mode->driver.vbeline.crtc);
	s += sprintf(s,")");
}

void vbeline_wait_vsync(void) {
	if (vbe_mode_active())
		vbe_wait_vsync();
	else
		vga_wait_vsync();
}

int vbeline_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	if (vbe_mode_active())
		return vbe_palette8_set(palette,start,count,waitvsync);
	else {
		vga_palette8_set(palette,start,count,waitvsync);
		return 0;
	}
}
