/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2001 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __VBELINE_H
#define __VBELINE_H

#include "screen.h"
#include "../card/card.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Driver capability */
#define VBELINE_CAP_SINGLESCAN 0x1 /* Single scan mode */
#define VBELINE_CAP_DOUBLESCAN 0x2 /* Double scan mode */
#define VBELINE_CAP_INTERLACE 0x4 /* Interlaced modes */
#define VBELINE_CAP_8BIT 0x10 /* 8 bit modes */
#define VBELINE_CAP_15BIT 0x20 /* 15 bit modes */
#define VBELINE_CAP_16BIT 0x40 /* 16 bit modes */
#define VBELINE_CAP_24BIT 0x80 /* 24 bit modes */
#define VBELINE_CAP_32BIT 0x100	/* 32 bit modes */
#define VBELINE_CAP_REQUIRES_VBE3_SET 0x10000 /* Require a vbe3 mode set */
#define VBELINE_CAP_REQUIRES_VBE_SAVE 0x20000 /* Require a mode save/restore with the vbe services */

typedef struct vbeline_driver_struct {
	const char* name; /* Driver name */
	unsigned cap; /* Driver capability */
	int type; /* VIDEO_VBELINE_DRIVER_* */
	int (*detect)(void);
	int (*set)(const card_crtc*);
	void (*reset)(void);
} vbeline_driver;

typedef struct vbeline_internal_struct {
	int active; /* !=0 if present */
	vbeline_driver* driver;

	int mode_active; /* !=0 if mode set */

	unsigned size_x;
	unsigned size_y;
	unsigned virtual_x;
	unsigned virtual_y;

	/* requested mode */
	unsigned mode_request;
	vbe_CRTCInfoBlock vbe3_crtc_request;
	vbe_ModeInfoBlock vbe_info_request;
	video_crtc crtc_request;

	void* state_ptr;
	unsigned state_size;
} vbeline_internal;

vbeline_internal vbeline_state;

static __inline__ int vbeline_active(void) {
	return vbeline_state.active != 0;
}

static __inline__ int vbeline_mode_active(void) {
	return vbeline_state.mode_active != 0;
}

static __inline__ const char* vbeline_driver_name(void) {
	return vbeline_state.driver->name;
}

static __inline__ unsigned vbeline_driver_capability(void) {
	assert(	vbeline_active() );
	return vbeline_state.driver->cap;
}

/* Return the selector for accessing in writing the video memory */
extern unsigned FAST (*vbeline_write_selector)(void);

/* Return the offset for accessing in writing the video memory */
extern unsigned FAST (*vbeline_write_line)(unsigned y);


int vbeline_pixelclock_getnext(unsigned* pixelclock, unsigned mode);
int vbeline_pixelclock_getpred(unsigned* pixelclock, unsigned mode);

void vbeline_wait_vsync(void);
int vbeline_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync);

int vbeline_mode_import(video_mode* mode);

void vbeline_init(void);
void vbeline_done(void);
int vbeline_mode_set(vbeline_video_mode* mode);
void vbeline_mode_done(void);

video_error vbeline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode);
video_error vbeline_mode_generate(video_mode* mode, const char* name, const video_crtc* crtc, int bits);
void video_mode_container_insert_vbeline(video_mode_container* vmc);
void video_mode_vbeline_print(char* s, const video_mode* mode);

#ifdef __cplusplus
}
#endif


#endif
