#ifndef __SCRIPT_H
#define __SCRIPT_H

/***************************************************************************/
/* Expression */

struct script_exp;

struct script_exp_op1e {
	struct script_exp* arg0;
};

struct script_exp_op2ee {
	struct script_exp* arg0;
	struct script_exp* arg1;
};

struct script_exp_op1v {
	int arg0;
};

struct script_exp_op1s {
	const char* arg0;
};

struct script_exp_op1f {
	const char* arg0;
};

struct script_exp_op2fe {
	const char* arg0;
	struct script_exp* arg1;
};

struct script_exp_op3fee {
	const char* arg0;
	struct script_exp* arg1;
	struct script_exp* arg2;
};

union script_exp_data {
	struct script_exp_op2ee op2ee;
	struct script_exp_op1e op1e;
	struct script_exp_op1v op1v;
	struct script_exp_op1s op1s;
	struct script_exp_op1f op1f;
	struct script_exp_op2fe op2fe;
	struct script_exp_op3fee op3fee;
};

/* Script expression */
#define SCRIPT_EXP_EXPRESSION 0
#define SCRIPT_EXP_VALUE 1
#define SCRIPT_EXP_VARIABLE 2
#define SCRIPT_EXP_ADD 3
#define SCRIPT_EXP_SUB 4
#define SCRIPT_EXP_AND 5
#define SCRIPT_EXP_OR 6
#define SCRIPT_EXP_XOR 7
#define SCRIPT_EXP_NOT 8
#define SCRIPT_EXP_L 9
#define SCRIPT_EXP_G 10
#define SCRIPT_EXP_E 11
#define SCRIPT_EXP_LE 12
#define SCRIPT_EXP_GE 13
#define SCRIPT_EXP_F0 14
#define SCRIPT_EXP_F1 15
#define SCRIPT_EXP_F2 16
#define SCRIPT_EXP_SL 17
#define SCRIPT_EXP_SR 18
#define SCRIPT_EXP_LOR 19
#define SCRIPT_EXP_LAND 20
#define SCRIPT_EXP_LNOT 21

struct script_exp {
	int type;
	union script_exp_data data;
};

int script_evaluate(const struct script_exp* exp);

/* Symbol callback */
int script_symbol_check(const char* sym);
int script_symbol_get(const char* sym);
void script_symbol_set(const char* sym, int arg0);
int script_function_check(const char* sym, int argc);
int script_function0_get(const char* sym);
int script_function1_get(const char* sym, int arg0);
int script_function2_get(const char* sym, int arg0, int arg1);

/***************************************************************************/
/* Parse */

struct script_exp* script_exp_make_op1v(int type, int arg0);
struct script_exp* script_exp_make_op1s(int type, const char* arg0);
struct script_exp* script_exp_make_op1e(int type, struct script_exp* arg0);
struct script_exp* script_exp_make_op2ee(int type, struct script_exp* arg0, struct script_exp* arg1);
struct script_exp* script_exp_make_op1f(int type, const char* arg0);
struct script_exp* script_exp_make_op2fe(int type, const char* arg0, struct script_exp* arg1);
struct script_exp* script_exp_make_op3fee(int type, const char* arg0, struct script_exp* arg1, struct script_exp* arg2);

/***************************************************************************/
/* Commands */

/* Script commands */
#define SCRIPT_CMD_WAIT 1
#define SCRIPT_CMD_LOOP 2
#define SCRIPT_CMD_INNER 3
#define SCRIPT_CMD_EVALUATE 4
#define SCRIPT_CMD_WHILE 5
#define SCRIPT_CMD_IF 6
#define SCRIPT_CMD_DELAY 7
#define SCRIPT_CMD_REPEAT 8

/* Commands data */
struct script_cmd_op1e {
	struct script_exp* arg0;
};

struct script_cmd_op1ed {
	struct script_exp* arg0;
	int value;
	int value_set;
};

struct script_cmd_op1c {
	struct script_cmd* arg0;
};

struct script_cmd_op2ec {
	struct script_exp* arg0;
	struct script_cmd* arg1;
};

struct script_cmd_op2ecd {
	struct script_exp* arg0;
	struct script_cmd* arg1;
	int value;
	int value_set;
};

union script_cmd_data {
	struct script_cmd_op1e op1e;
	struct script_cmd_op1ed op1ed;
	struct script_cmd_op1c op1c;
	struct script_cmd_op2ec op2ec;
	struct script_cmd_op2ecd op2ecd;
};

/***************************************************************************/
/* Script */

struct script_cmd {
	struct script_cmd* next; /* ==0 if none */
	int type;
	union script_cmd_data data;
};

/***************************************************************************/
/* Parse */

struct script_cmd* script_cmd_make_op1c(struct script_cmd* arg0);
struct script_cmd* script_cmd_make_op2sc(const char* arg0, struct script_cmd* arg1);
struct script_cmd* script_cmd_make_op2se(const char* arg0, struct script_exp* arg1);
struct script_cmd* script_cmd_make_op2cc(struct script_cmd* arg0, struct script_cmd* arg1);
struct script_cmd* script_cmd_make_op3sec(const char* arg0, struct script_exp* arg1, struct script_cmd* arg2);

/***************************************************************************/
/* State */

#define SCRIPT_STATE_STACK_MAX 16

struct script_state {
	struct script_cmd* cursor_map[SCRIPT_STATE_STACK_MAX];
	unsigned cursor_mac;
	unsigned time_to_play;
};

/* Port callback */
unsigned char script_port_read(int address);
void script_port_write(int address,unsigned char value);

/* Error callback */
void script_error(const char* s);

struct script_cmd* script_parse(const char* text);
void script_free(struct script_cmd* script);

struct script_state* script_run_alloc(void);
void script_run_free(struct script_state* state);
void script_run_restart(struct script_state* state,struct script_cmd* cmd);
int script_run_end(const struct script_state* state);

/* Unit time (1 second) for the idle call */
#define SCRIPT_TIME_UNIT 1000000

/* Unit time (1 second) for the script delay() call */
#define SCRIPT_DELAY_UNIT 1000

void script_run(struct script_state* state, unsigned time_to_play);

#endif
